# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from pyimpspec import (
    Circuit,
    ParsingError,
)
from typing import (
    Callable,
    List,
    Optional,
)
from numpy import (
    allclose,
    array,
    ndarray,
)
import pyimpspec
import dearpygui.dearpygui as dpg
from deareis.gui.plots import Nyquist
import deareis.themes as themes
from deareis.utility import calculate_window_position_dimensions
from deareis.tooltips import attach_tooltip
import deareis.tooltips as tooltips
from deareis.gui.circuit_editor import CircuitEditor
from deareis.signals import Signal
import deareis.signals as signals
from deareis.data import (
    DataSet,
    FitResult,
)
from deareis.keybindings import (
    is_alt_down,
    is_control_down,
)


def format_fit_labels(fits: List[FitResult]) -> List[str]:
    old_labels: List[str] = list(map(lambda _: _.get_label(), fits))
    longest_cdc: int = max(list(map(lambda _: len(_[: _.find(" ")]), old_labels)) + [1])
    new_labels: List[str] = []
    old_key: str
    for old_key in old_labels:
        cdc, timestamp = (
            old_key[: old_key.find(" ")],
            old_key[old_key.find(" ") + 1 :],
        )
        new_labels.append(f"{cdc.ljust(longest_cdc)} {timestamp}")
    return new_labels


class SubtractImpedance:
    def __init__(
        self,
        data: DataSet,
        data_sets: List[DataSet],
        fits: List[FitResult],
        callback: Callable,
    ):
        assert type(data) is DataSet, data
        assert type(data_sets) is list and all(
            map(lambda _: type(_) is DataSet, data_sets)
        ), data_sets
        self.data: DataSet = data
        self.data_sets: List[DataSet] = [
            _
            for _ in sorted(data_sets, key=lambda _: _.get_label())
            if _ != data
            and data.get_num_points(masked=None) == _.get_num_points(masked=None)
            and allclose(data.get_frequency(masked=None), _.get_frequency(masked=None))
        ]
        self.data_labels: List[str] = list(map(lambda _: _.get_label(), self.data_sets))
        self.fits: List[FitResult] = fits
        self.fit_labels: List[str] = format_fit_labels(fits)
        self.preview_data: DataSet = DataSet.from_dict(data.to_dict())
        self.callback: Callable = callback
        self.options: List[str] = [
            "Constant:",
            " Circuit:",
            "     Fit:",
            "Spectrum:",
        ]
        self.circuit_editor_window: int = -1
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            label="Subtract impedance",
            modal=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            tag=self.window,
            on_close=self.close,
        ):
            self.preview_window: int = dpg.generate_uuid()
            with dpg.child_window(border=False, tag=self.preview_window):
                with dpg.child_window(
                    width=-1,
                    height=104,
                ):
                    self.radio_buttons: int = dpg.generate_uuid()
                    with dpg.group(horizontal=True):
                        dpg.add_radio_button(
                            items=self.options,
                            default_value=self.options[0],
                            callback=self.select_option,
                            tag=self.radio_buttons,
                        )
                        with dpg.group():
                            self.constant_group: int = dpg.generate_uuid()
                            with dpg.group(horizontal=True, tag=self.constant_group):
                                dpg.add_text("Z' = ")
                                self.constant_real: int = dpg.generate_uuid()
                                dpg.add_input_float(
                                    label="ohm,",
                                    default_value=0.0,
                                    step=0.0,
                                    format="%.3g",
                                    on_enter=True,
                                    width=100,
                                    tag=self.constant_real,
                                    callback=self.update_preview,
                                )
                                dpg.add_text('-Z" = ')
                                self.constant_imag: int = dpg.generate_uuid()
                                dpg.add_input_float(
                                    label="ohm",
                                    default_value=0.0,
                                    step=0.0,
                                    format="%.3g",
                                    on_enter=True,
                                    width=100,
                                    tag=self.constant_imag,
                                    callback=self.update_preview,
                                )
                            self.circuit_group: int = dpg.generate_uuid()
                            with dpg.group(horizontal=True, tag=self.circuit_group):
                                self.circuit_cdc: int = dpg.generate_uuid()
                                dpg.add_input_text(
                                    hint="Input CDC",
                                    on_enter=True,
                                    width=314,
                                    tag=self.circuit_cdc,
                                    callback=self.update_preview,
                                )
                                dpg.add_button(
                                    label="Edit",
                                    callback=self.edit_circuit,
                                )
                                attach_tooltip(tooltips.general.open_circuit_editor)
                            self.fit_group: int = dpg.generate_uuid()
                            with dpg.group(horizontal=True, tag=self.fit_group):
                                self.fit_combo: int = dpg.generate_uuid()
                                dpg.add_combo(
                                    items=self.fit_labels,
                                    default_value=self.fit_labels[0]
                                    if self.fit_labels
                                    else "",
                                    width=358,
                                    tag=self.fit_combo,
                                    callback=self.update_preview,
                                )
                            self.spectrum_group: int = dpg.generate_uuid()
                            with dpg.group(horizontal=True, tag=self.spectrum_group):
                                self.spectrum_combo: int = dpg.generate_uuid()
                                dpg.add_combo(
                                    items=self.data_labels,
                                    default_value=self.data_labels[0]
                                    if self.data_labels
                                    else "",
                                    width=358,
                                    tag=self.spectrum_combo,
                                    callback=self.update_preview,
                                )
                self.nyquist_plot: Nyquist = Nyquist(width=-1, height=-24)
                self.nyquist_plot.plot(
                    real=array([]),
                    imaginary=array([]),
                    label="Before",
                    theme=themes.nyquist.data,
                    show_label=False,
                )
                self.nyquist_plot.plot(
                    real=array([]),
                    imaginary=array([]),
                    label="Before",
                    line=True,
                    theme=themes.nyquist.data,
                )
                self.nyquist_plot.plot(
                    real=array([]),
                    imaginary=array([]),
                    label="After",
                    theme=themes.bode.phase_data,
                    show_label=False,
                )
                self.nyquist_plot.plot(
                    real=array([]),
                    imaginary=array([]),
                    label="After",
                    line=True,
                    theme=themes.bode.phase_data,
                )
                dpg.add_button(
                    label="Accept",
                    callback=self.accept,
                )
            self.circuit_editor_window = dpg.generate_uuid()
            with dpg.child_window(
                border=False,
                show=False,
                tag=self.circuit_editor_window,
            ):
                self.circuit_editor: CircuitEditor = CircuitEditor(
                    window=self.circuit_editor_window,
                    callback=self.accept_circuit,
                )
        self.key_handler: int = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.close,
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Return,
                callback=lambda: self.accept(keybinding=True),
            )
        self.select_option(self.radio_buttons, self.options[0])

    def close(self):
        if not dpg.is_item_visible(self.constant_real):
            return
        self.circuit_editor.hide()
        dpg.hide_item(self.window)
        dpg.delete_item(self.window)
        dpg.delete_item(self.key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    def accept(self, keybinding: bool = False):
        if not dpg.is_item_visible(self.constant_real):
            return
        elif (
            self.circuit_editor_window > 0
            and dpg.does_item_exist(self.circuit_editor_window)
            and dpg.is_item_shown(self.circuit_editor_window)
        ):
            return
        elif keybinding is True and not (
            is_control_down()
            if dpg.get_platform() == dpg.mvPlatform_Windows
            else is_alt_down()
        ):
            return
        self.close()
        self.callback(DataSet.from_dict(self.preview_data.to_dict()))

    def select_option(self, sender: int, value: str):
        item_type: str

        def disable_group(group: int):
            for item in dpg.get_item_children(group, slot=1):
                item_type = dpg.get_item_type(item)
                if item_type.endswith("mvText") or item_type.endswith("mvTooltip"):
                    continue
                dpg.disable_item(item)

        def enable_group(group: int):
            for item in dpg.get_item_children(group, slot=1):
                item_type = dpg.get_item_type(item)
                if item_type.endswith("mvText") or item_type.endswith("mvTooltip"):
                    continue
                dpg.enable_item(item)

        index: int = self.options.index(value)
        if index == 0:
            enable_group(self.constant_group)
            disable_group(self.circuit_group)
            disable_group(self.fit_group)
            disable_group(self.spectrum_group)
        elif index == 1:
            disable_group(self.constant_group)
            enable_group(self.circuit_group)
            disable_group(self.fit_group)
            disable_group(self.spectrum_group)
        elif index == 2:
            disable_group(self.constant_group)
            disable_group(self.circuit_group)
            enable_group(self.fit_group)
            disable_group(self.spectrum_group)
        elif index == 3:
            disable_group(self.constant_group)
            disable_group(self.circuit_group)
            disable_group(self.fit_group)
            enable_group(self.spectrum_group)
        else:
            raise Exception(f"Unsupported option: {value}")
        self.update_preview()

    def update_preview(self):
        index: int = self.options.index(dpg.get_value(self.radio_buttons))
        f: ndarray = self.data.get_frequency(masked=None)
        Z: ndarray = self.data.get_impedance(masked=None)
        if index == 0:
            Z_const: complex = complex(
                dpg.get_value(self.constant_real),
                -dpg.get_value(self.constant_imag),
            )
            Z = Z - Z_const
        elif index == 1:
            try:
                circuit: Circuit = pyimpspec.parse_cdc(dpg.get_value(self.circuit_cdc))
            except Exception:
                return
            Z = Z - circuit.impedances(f)
        elif index == 2:
            if len(self.fits) > 0:
                fit: FitResult
                fit = self.fits[self.fit_labels.index(dpg.get_value(self.fit_combo))]
                Z = Z - fit.circuit.impedances(f)
        elif index == 3:
            if len(self.data_sets) > 0:
                spectrum: DataSet
                spectrum = self.data_sets[
                    self.data_labels.index(dpg.get_value(self.spectrum_combo))
                ]
                Z = Z - spectrum.get_impedance(masked=None)
        else:
            raise Exception("Unsupported option!")
        dictionary: dict = self.preview_data.to_dict()
        dictionary.update(
            {
                "real": list(Z.real),
                "imaginary": list(Z.imag),
            }
        )
        self.preview_data = DataSet.from_dict(dictionary)
        self.update_plot()

    def update_plot(self):
        real: ndarray
        imag: ndarray
        real, imag = self.data.get_nyquist_data(masked=None)
        self.nyquist_plot.update(
            index=0,
            real=real,
            imaginary=imag,
        )
        self.nyquist_plot.update(
            index=1,
            real=real,
            imaginary=imag,
        )
        real, imag = self.preview_data.get_nyquist_data(masked=None)
        self.nyquist_plot.update(
            index=2,
            real=real,
            imaginary=imag,
        )
        self.nyquist_plot.update(
            index=3,
            real=real,
            imaginary=imag,
        )
        self.nyquist_plot.queue_limits_adjustment()

    def edit_circuit(self):
        dpg.hide_item(self.preview_window)
        circuit: Optional[Circuit] = None
        try:
            circuit = pyimpspec.parse_cdc(dpg.get_value(self.circuit_cdc) or "[]")
        except ParsingError:
            pass
        self.circuit_editor.show(circuit)

    def accept_circuit(self, circuit: Optional[Circuit]):
        self.circuit_editor.hide()
        dpg.show_item(self.preview_window)
        self.update_cdc(circuit)

    def update_cdc(self, circuit: Optional[Circuit]):
        if circuit is not None:
            for element in circuit.get_elements():
                element.set_label("")
                for param in element.get_parameters():
                    element.set_fixed(param, True)
        assert dpg.does_item_exist(self.circuit_cdc)
        dpg.set_value(
            self.circuit_cdc,
            circuit.to_string(6) if circuit is not None else "",
        )
        dpg.show_item(self.preview_window)
        dpg.split_frame(delay=33)
        self.update_preview()



from setuptools import (
    setup,
    find_packages,
)
from os import walk
from os.path import (
    basename,
    exists,
    join,
)


def update_file(src: str, dst: str):
    if not exists(src):
        return
    src_contents = ""
    with open(src, "r") as fp:
        src_contents = fp.read()
    if exists(dst):
        with open(dst, "r") as fp:
            if fp.read() == src_contents:
                return
    with open(dst, "w") as fp:
        fp.write(src_contents)


entry_points = {
    "gui_scripts": [
        "deareis = deareis.program:main",
    ],
    "console_scripts": [
        "deareis-debug = deareis.program:debug",  # For the convenience of users on Windows
    ],
}

dependencies = [
    "dearpygui==1.7.3",  # Used to implement the GUI.
    "pyimpspec>=3.2.3",  # Used for parsing, fitting, and analyzing impedance spectra.
    "requests>=2.28.1",  # Used to check package status on PyPI.
    "xdg>=5.1.1",  # Used to figure out where to place config, state, etc. files.
]

dev_dependencies = [
    "flake8",
    "setuptools",
]

optional_dependencies = {
    "cvxpy": "cvxpy>=1.2.1",  # Used in the DRT calculations (TR-RBF method)
    "kvxopt": "kvxopt>=1.3.0",  # Fork of cvxopt that may provide wheels for additional platforms
    "dev": dev_dependencies,
}

licenses = []
for _, _, files in walk("LICENSES"):
    licenses.extend(
        list(
            map(
                lambda _: join("LICENSES", _),
                filter(lambda _: _.startswith("LICENSE-"), files),
            )
        )
    )

data_files = [
    "COPYRIGHT",
    "CONTRIBUTORS",
    "LICENSES/README.md",
    "src/deareis/gui/changelog/CHANGELOG.md",
] + licenses

# The version number defined below is propagated to /src/deareis/version.py
# when running this script.
version = "3.4.2"

if __name__ == "__main__":
    with open("requirements.txt", "w") as fp:
        fp.write("\n".join(dependencies))
    with open("dev-requirements.txt", "w") as fp:
        fp.write("\n".join(dev_dependencies))
    with open("version.txt", "w") as fp:
        fp.write(version)
    assert version.strip != ""
    copyright_notice = ""
    with open("COPYRIGHT") as fp:
        copyright_notice = fp.read().strip()
    assert copyright_notice.strip() != ""
    with open(join("src", "deareis", "version.py"), "w") as fp:
        fp.write(f'{copyright_notice}\n\nPACKAGE_VERSION: str = "{version}"')
    # The changelog bundled with the package will also be updated when running this script.
    update_file(
        join("CHANGELOG.md"),
        join("src", "deareis", "gui", "changelog", "CHANGELOG.md"),
    )
    # The licenses bundled with the package will also be updated when running this script.
    update_file(
        join("LICENSE"),
        join("src", "deareis", "gui", "licenses", "LICENSE-DearEIS.txt"),
    )
    list(
        map(
            lambda _: update_file(
                _,
                join("src", "deareis", "gui", "licenses", basename(_)),
            ),
            licenses,
        )
    )
    setup(
        name="deareis",
        version=version,
        author="DearEIS developers",
        packages=find_packages(where="src"),
        package_dir={"": "src"},
        include_package_data=True,
        data_files=data_files,
        url="https://vyrjana.github.io/DearEIS",
        project_urls={
            "Documentation": "https://vyrjana.github.io/DearEIS/api/",
            "Source Code": "https://github.com/vyrjana/DearEIS",
            "Bug Tracker": "https://github.com/vyrjana/DearEIS/issues",
        },
        license="GPLv3",
        description="A GUI program for analyzing, simulating, and visualizing impedance spectra.",
        long_description=open("README.md").read(),
        long_description_content_type="text/markdown",
        entry_points=entry_points,
        install_requires=dependencies,
        extras_require=optional_dependencies,
        python_requires=">=3.8",
        classifiers=[
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Operating System :: MacOS",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX :: Linux",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3 :: Only",
            "Topic :: Scientific/Engineering :: Chemistry",
            "Topic :: Scientific/Engineering :: Physics",
            "Topic :: Scientific/Engineering",
        ],
    )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import Callable, Dict, List
from numpy import array, ndarray
import dearpygui.dearpygui as dpg
from deareis.gui.plots import Nyquist
import deareis.themes as themes
from deareis.data import DataSet
from deareis.utility import (
    align_numbers,
    calculate_window_position_dimensions,
    format_number,
)
from deareis.signals import Signal
import deareis.signals as signals
from deareis.tooltips import attach_tooltip
from deareis.keybindings import (
    is_alt_down,
    is_control_down,
)


class ToggleDataPoints:
    def __init__(self, data: DataSet, callback: Callable):
        assert type(data) is DataSet
        self.data: DataSet = data
        self.callback: Callable = callback
        self.preview_data: DataSet = DataSet.from_dict(data.to_dict())
        self.original_mask: Dict[int, bool] = data.get_mask()
        self.final_mask: Dict[int, bool] = {
            i: state for i, state in self.original_mask.items()
        }
        self.labels: List[str] = []
        indices: List[str] = align_numbers(
            list(map(str, range(1, data.get_num_points(masked=None) + 1)))
        )
        freq: List[str] = list(
            map(
                lambda _: _.ljust(9),
                align_numbers(
                    list(
                        map(
                            lambda _: format_number(_, 1, 9),
                            data.get_frequency(masked=None),
                        )
                    )
                ),
            )
        )
        Z: ndarray = data.get_impedance(masked=None)
        real: List[str] = list(
            map(
                lambda _: _.ljust(10),
                align_numbers(list(map(lambda _: format_number(_, 1, 10), Z.real))),
            )
        )
        imag: List[str] = list(
            map(
                lambda _: _.ljust(10),
                align_numbers(list(map(lambda _: format_number(_, 1, 10), -Z.imag))),
            )
        )
        i: str
        f: str
        re: str
        im: str
        for (i, f, re, im) in zip(
            indices,
            freq,
            real,
            imag,
        ):
            self.labels.append(
                f"{i}: " + f"f = {f} | " + f"Z'= {re} | " + f'-Z" = {im}'
            )

        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            label="Toggle points",
            modal=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            tag=self.window,
            on_close=self.close,
        ):
            with dpg.group(horizontal=True):
                dpg.add_text("From")
                self.from_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    default_value=self.labels[0],
                    width=-1,
                    callback=self.update_items,
                    tag=self.from_combo,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("  To")
                self.to_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    default_value=self.labels[-1],
                    width=-1,
                    callback=self.update_items,
                    tag=self.to_combo,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("  ?")
                attach_tooltip(
                    """Points can also be chosen by drawing rectangle while holding down the middle-mouse button.""".strip()
                )
                dpg.add_button(
                    label="Exclude all",
                    callback=self.exclude,
                )
                dpg.add_button(
                    label="Include all",
                    callback=self.include,
                )
            self.nyquist_plot: Nyquist = Nyquist(width=-1, height=-24)
            self.nyquist_plot.plot(
                real=array([]),
                imaginary=array([]),
                label="Excluded",
                theme=themes.nyquist.data,
            )
            real, imag = self.preview_data.get_nyquist_data(masked=False)
            self.nyquist_plot.plot(
                real=array([]),
                imaginary=array([]),
                label="Included",
                theme=themes.bode.phase_data,
                show_label=False,
            )
            self.nyquist_plot.plot(
                real=array([]),
                imaginary=array([]),
                label="Included",
                line=True,
                theme=themes.bode.phase_data,
            )
            dpg.configure_item(self.nyquist_plot._plot, query=True)
            dpg.add_button(
                label="Accept",
                callback=self.accept,
            )
        self.key_handler: int = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.close,
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Return,
                callback=lambda: self.accept(keybinding=True),
            )
        self.update_items(self.from_combo, self.labels[0])
        self.update_items(self.to_combo, self.labels[-1])
        self.update_preview()

    def update_preview(self):
        real: ndarray
        imag: ndarray
        real, imag = self.preview_data.get_nyquist_data(masked=True)
        self.nyquist_plot.update(
            index=0,
            real=real,
            imaginary=imag,
        )
        real, imag = self.preview_data.get_nyquist_data(masked=False)
        self.nyquist_plot.update(
            index=1,
            real=real,
            imaginary=imag,
        )
        self.nyquist_plot.update(
            index=2,
            real=real,
            imaginary=imag,
        )
        self.nyquist_plot.queue_limits_adjustment()

    def close(self):
        dpg.hide_item(self.window)
        dpg.delete_item(self.window)
        dpg.delete_item(self.key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    def accept(self, keybinding: bool = False):
        if keybinding is True and not (
            is_control_down()
            if dpg.get_platform() == dpg.mvPlatform_Windows
            else is_alt_down()
        ):
            return
        if dpg.is_plot_queried(self.nyquist_plot._plot):
            sx, ex, sy, ey = dpg.get_plot_query_area(self.nyquist_plot._plot)
            for i, Z in enumerate(self.data.get_impedance(masked=None)):
                if Z.real >= sx and Z.real <= ex and -Z.imag >= sy and -Z.imag <= ey:
                    self.final_mask[i] = not self.final_mask[i]
        self.callback(self.final_mask)
        self.close()

    def exclude(self):
        self.final_mask = {i: True for i in self.original_mask}
        self.preview_data.set_mask(self.final_mask)
        self.update_preview()

    def include(self):
        self.final_mask = {i: False for i in self.original_mask}
        self.preview_data.set_mask(self.final_mask)
        self.update_preview()

    def update_items(self, sender: int, label: str):
        index: int = self.labels.index(label)
        receiver: int
        items: List[str]
        if sender == self.from_combo:
            receiver = self.to_combo
            items = self.labels[index + 1 :]
        elif sender == self.to_combo:
            receiver = self.from_combo
            items = self.labels[:index]
        dpg.configure_item(receiver, items=items)
        start: int = self.labels.index(dpg.get_value(self.from_combo))
        end: int = self.labels.index(dpg.get_value(self.to_combo))
        assert end > start
        self.final_mask = {}
        for i, state in self.original_mask.items():
            if i >= start and i <= end:
                self.final_mask[i] = not state
            else:
                self.final_mask[i] = state
        self.preview_data.set_mask(self.final_mask)
        self.update_preview()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from uuid import uuid4
from time import time
import dearpygui.dearpygui as dpg
import deareis.themes as themes
from deareis.themes import PLOT_MARKERS
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
)
from deareis.utility import calculate_window_position_dimensions
from deareis.tooltips import attach_tooltip
from deareis.state import STATE
from deareis.gui.plots import (
    Bode,
    DRT,
    Impedance,
    MuXps,
    Nyquist,
    Residuals,
)
from numpy import (
    array,
    exp,
    log10 as log,
    logspace,
    ndarray,
)
from numpy.random import normal
from pyimpspec import Circuit
from deareis.data import (
    DRTResult,
    DRTSettings,
    DataSet,
)
from deareis.enums import (
    DRTMethod,
    DRTMode,
    RBFShape,
    RBFType,
)
import pyimpspec
from deareis.signals import Signal
import deareis.signals as signals


# TODO: Refactor color and marker widgets to reduce code duplication
# TODO: Refactor update_*_* functions to reduce code duplication


class AppearanceSettings:
    def __init__(self):
        self.marker_items: List[str] = list(PLOT_MARKERS.keys())
        self.marker_label_lookup: Dict[int, str] = {
            v: k for k, v in PLOT_MARKERS.items()
        }
        self.label_pad: int = 56
        self.data: DataSet
        self.sim_data: DataSet
        self.smooth_data: DataSet
        self.real_residual: ndarray
        self.imag_residual: ndarray
        self.noise: ndarray
        self.drt: DRTResult
        (
            self.data,
            self.sim_data,
            self.smooth_data,
            self.real_residual,
            self.imag_residual,
            self.noise,
            self.drt,
        ) = self.generate_data()
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions(600, 540)
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            label="Settings - appearance",
            modal=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            no_move=False,
            no_resize=True,
            on_close=self.close_window,
            tag=self.window,
        ):
            with dpg.collapsing_header(label="General", default_open=True):
                with dpg.group(horizontal=True):
                    dpg.add_text(
                        "Number of points per decade in simulated response".rjust(
                            self.label_pad
                        )
                    )
                    attach_tooltip(
                        """
This affects how smooth the lines will look when plotting the impedance response of, e.g., fitted circuits, but it may also affect performance when rendering the graphical user interface. This setting also affects how many points are included when copying plot data as character-separated values (CSV). Changes made to this setting will take effect the next time a plot is redrawn.
                        """.strip()
                    )
                    dpg.add_slider_int(
                        default_value=STATE.config.num_per_decade_in_simulated_lines,
                        min_value=1,
                        max_value=200,
                        clamped=True,
                        callback=self.update_simulated_num_per_decade,
                        width=-1,
                    )
                dpg.add_spacer(height=8)
            with dpg.collapsing_header(label="Bode plots", default_open=True):
                self.bode_plot = Bode(width=-1, height=200)
                self.bode_data_mag_color: int = dpg.generate_uuid()
                self.bode_data_phase_color: int = dpg.generate_uuid()
                self.bode_sim_mag_color: int = dpg.generate_uuid()
                self.bode_sim_phase_color: int = dpg.generate_uuid()
                self.bode_data_mag_marker: int = dpg.generate_uuid()
                self.bode_data_phase_marker: int = dpg.generate_uuid()
                self.bode_sim_mag_marker: int = dpg.generate_uuid()
                self.bode_sim_phase_marker: int = dpg.generate_uuid()
                # Data colors and markers
                with dpg.group(horizontal=True):
                    dpg.add_text("Data - magnitude".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["bode_magnitude_data"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_bode_color,
                        user_data=themes.bode.magnitude_data,
                        tag=self.bode_data_mag_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["bode_magnitude_data"]
                        ],
                        callback=self.update_bode_marker,
                        user_data=themes.bode.magnitude_data,
                        tag=self.bode_data_mag_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Data - phase".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["bode_phase_data"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_bode_color,
                        user_data=themes.bode.phase_data,
                        tag=self.bode_data_phase_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["bode_phase_data"]
                        ],
                        callback=self.update_bode_marker,
                        user_data=themes.bode.phase_data,
                        tag=self.bode_data_phase_marker,
                        width=-1,
                    )
                # Sim/fit colors and markers
                with dpg.group(horizontal=True):
                    dpg.add_text("Fit/simulation - magnitude".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["bode_magnitude_simulation"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_bode_color,
                        user_data=themes.bode.magnitude_simulation,
                        tag=self.bode_sim_mag_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["bode_magnitude_simulation"]
                        ],
                        callback=self.update_bode_marker,
                        user_data=themes.bode.magnitude_simulation,
                        tag=self.bode_sim_mag_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Fit/simulation - phase".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["bode_phase_simulation"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_bode_color,
                        user_data=themes.bode.phase_simulation,
                        tag=self.bode_sim_phase_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["bode_phase_simulation"]
                        ],
                        callback=self.update_bode_marker,
                        user_data=themes.bode.phase_simulation,
                        tag=self.bode_sim_phase_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("".rjust(self.label_pad))
                    dpg.add_button(
                        label="Restore defaults",
                        callback=self.reset_bode_plot,
                        width=-1,
                    )
                self.update_bode_plot(True)
                dpg.add_spacer(height=8)
            with dpg.collapsing_header(label="DRT plots", default_open=True):
                self.drt_plot = DRT(width=-1, height=200)
                self.drt_real_gamma_color: int = dpg.generate_uuid()
                self.drt_imaginary_gamma_color: int = dpg.generate_uuid()
                self.drt_mean_gamma_color: int = dpg.generate_uuid()
                self.drt_credible_intervals_color: int = dpg.generate_uuid()
                with dpg.group(horizontal=True):
                    dpg.add_text("Gamma/real".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["drt_real_gamma"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_drt_color,
                        user_data=themes.drt.real_gamma,
                        tag=self.drt_real_gamma_color,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Imaginary".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["drt_imaginary_gamma"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_drt_color,
                        user_data=themes.drt.imaginary_gamma,
                        tag=self.drt_imaginary_gamma_color,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Mean and 3-sigma CI".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["drt_mean_gamma"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_drt_color,
                        user_data=themes.drt.mean_gamma,
                        tag=self.drt_mean_gamma_color,
                    )
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["drt_credible_intervals"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_drt_color,
                        user_data=themes.drt.credible_intervals,
                        tag=self.drt_credible_intervals_color,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("".rjust(self.label_pad))
                    dpg.add_button(
                        label="Restore defaults",
                        callback=self.reset_drt_plot,
                        width=-1,
                    )
                self.update_drt_plot(True)
                dpg.add_spacer(height=8)
            with dpg.collapsing_header(label="Impedance plots", default_open=True):
                self.impedance_plot = Impedance(width=-1, height=200)
                self.impedance_real_data_color: int = dpg.generate_uuid()
                self.impedance_real_simulation_color: int = dpg.generate_uuid()
                self.impedance_imaginary_data_color: int = dpg.generate_uuid()
                self.impedance_imaginary_simulation_color: int = dpg.generate_uuid()
                self.impedance_real_data_marker: int = dpg.generate_uuid()
                self.impedance_real_simulation_marker: int = dpg.generate_uuid()
                self.impedance_imaginary_data_marker: int = dpg.generate_uuid()
                self.impedance_imaginary_simulation_marker: int = dpg.generate_uuid()
                # Data colors and markers
                with dpg.group(horizontal=True):
                    dpg.add_text("Data - real".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["impedance_real_data"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_impedance_color,
                        user_data=themes.impedance.real_data,
                        tag=self.impedance_real_data_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["impedance_real_data"]
                        ],
                        callback=self.update_impedance_marker,
                        user_data=themes.impedance.real_data,
                        tag=self.impedance_real_data_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Data - imaginary".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["impedance_imaginary_data"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_impedance_color,
                        user_data=themes.impedance.imaginary_data,
                        tag=self.impedance_imaginary_data_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["impedance_imaginary_data"]
                        ],
                        callback=self.update_impedance_marker,
                        user_data=themes.impedance.imaginary_data,
                        tag=self.impedance_imaginary_data_marker,
                        width=-1,
                    )
                # Sim/fit colors and markers
                with dpg.group(horizontal=True):
                    dpg.add_text("Fit/simulation - real".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["impedance_real_simulation"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_impedance_color,
                        user_data=themes.impedance.real_simulation,
                        tag=self.impedance_real_simulation_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["impedance_real_simulation"]
                        ],
                        callback=self.update_impedance_marker,
                        user_data=themes.impedance.real_simulation,
                        tag=self.impedance_real_simulation_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Fit/simulation - imaginary".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors[
                            "impedance_imaginary_simulation"
                        ],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_impedance_color,
                        user_data=themes.impedance.imaginary_simulation,
                        tag=self.impedance_imaginary_simulation_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["impedance_imaginary_simulation"]
                        ],
                        callback=self.update_impedance_marker,
                        user_data=themes.impedance.imaginary_simulation,
                        tag=self.impedance_imaginary_simulation_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("".rjust(self.label_pad))
                    dpg.add_button(
                        label="Restore defaults",
                        callback=self.reset_impedance_plot,
                        width=-1,
                    )
                self.update_impedance_plot(True)
                dpg.add_spacer(height=8)
            with dpg.collapsing_header(label="Nyquist plots", default_open=True):
                self.nyquist_plot = Nyquist(width=-1, height=200)
                self.nyquist_data_color: int = dpg.generate_uuid()
                self.nyquist_sim_color: int = dpg.generate_uuid()
                self.nyquist_data_marker: int = dpg.generate_uuid()
                self.nyquist_sim_marker: int = dpg.generate_uuid()
                # Data colors and markers
                with dpg.group(horizontal=True):
                    dpg.add_text("Data".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["nyquist_data"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_nyquist_color,
                        user_data=themes.nyquist.data,
                        tag=self.nyquist_data_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["nyquist_data"]
                        ],
                        callback=self.update_nyquist_marker,
                        user_data=themes.nyquist.data,
                        tag=self.nyquist_data_marker,
                        width=-1,
                    )
                # Sim/fit colors and markers
                with dpg.group(horizontal=True):
                    dpg.add_text("Fit/simulation".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["nyquist_simulation"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_nyquist_color,
                        user_data=themes.nyquist.simulation,
                        tag=self.nyquist_sim_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["nyquist_simulation"]
                        ],
                        callback=self.update_nyquist_marker,
                        user_data=themes.nyquist.simulation,
                        tag=self.nyquist_sim_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("".rjust(self.label_pad))
                    dpg.add_button(
                        label="Restore defaults",
                        callback=self.reset_nyquist_plot,
                        width=-1,
                    )
                self.update_nyquist_plot(True)
                dpg.add_spacer(height=8)
            with dpg.collapsing_header(label="Residuals plots", default_open=True):
                self.residuals_plot = Residuals(width=-1, height=200)
                self.residuals_real_color: int = dpg.generate_uuid()
                self.residuals_imag_color: int = dpg.generate_uuid()
                self.residuals_real_marker: int = dpg.generate_uuid()
                self.residuals_imag_marker: int = dpg.generate_uuid()
                # Zre color and marker
                with dpg.group(horizontal=True):
                    dpg.add_text("Z' error".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["residuals_real"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_residuals_color,
                        user_data=themes.residuals.real,
                        tag=self.residuals_real_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["residuals_real"]
                        ],
                        callback=self.update_residuals_marker,
                        user_data=themes.residuals.real,
                        tag=self.residuals_real_marker,
                        width=-1,
                    )
                # Zim color and marker
                with dpg.group(horizontal=True):
                    dpg.add_text('Z" error'.rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["residuals_imaginary"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_residuals_color,
                        user_data=themes.residuals.imaginary,
                        tag=self.residuals_imag_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["residuals_imaginary"]
                        ],
                        callback=self.update_residuals_marker,
                        user_data=themes.residuals.imaginary,
                        tag=self.residuals_imag_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("".rjust(self.label_pad))
                    dpg.add_button(
                        label="Restore defaults",
                        callback=self.reset_residuals_plot,
                        width=-1,
                    )
                self.update_residuals_plot(True)
                dpg.add_spacer(height=8)
            with dpg.collapsing_header(label="µ-X² (pseudo) plots", default_open=True):
                self.muxps_plot = MuXps(width=-1, height=200)
                self.muxps_mu_criterion_color: int = dpg.generate_uuid()
                self.muxps_mu_color: int = dpg.generate_uuid()
                self.muxps_mu_highlight_color: int = dpg.generate_uuid()
                self.muxps_xps_color: int = dpg.generate_uuid()
                self.muxps_xps_highlight_color: int = dpg.generate_uuid()
                self.muxps_mu_marker: int = dpg.generate_uuid()
                self.muxps_xps_marker: int = dpg.generate_uuid()
                # Mu-criterion color
                with dpg.group(horizontal=True):
                    dpg.add_text("µ-criterion".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["mu_Xps_mu_criterion"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_muxps_color,
                        user_data=themes.mu_Xps.mu_criterion,
                        tag=self.muxps_mu_criterion_color,
                    )
                # Mu color and marker
                with dpg.group(horizontal=True):
                    dpg.add_text("µ".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["mu_Xps_mu"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_muxps_color,
                        user_data=themes.mu_Xps.mu,
                        tag=self.muxps_mu_color,
                    )
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["mu_Xps_mu_highlight"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_muxps_color,
                        user_data=themes.mu_Xps.mu_highlight,
                        tag=self.muxps_mu_highlight_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["mu_Xps_mu"]
                        ],
                        callback=self.update_muxps_marker,
                        user_data=themes.mu_Xps.mu,
                        tag=self.muxps_mu_marker,
                        width=-1,
                    )
                # Xps color and marker
                with dpg.group(horizontal=True):
                    dpg.add_text("X² (pseudo)".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["mu_Xps_Xps"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_muxps_color,
                        user_data=themes.mu_Xps.Xps,
                        tag=self.muxps_xps_color,
                    )
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["mu_Xps_Xps_highlight"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_muxps_color,
                        user_data=themes.mu_Xps.Xps_highlight,
                        tag=self.muxps_xps_highlight_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["mu_Xps_Xps"]
                        ],
                        callback=self.update_muxps_marker,
                        user_data=themes.mu_Xps.Xps,
                        tag=self.muxps_xps_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("".rjust(self.label_pad))
                    dpg.add_button(
                        label="Restore defaults",
                        callback=self.reset_muxps_plot,
                        width=-1,
                    )
                self.update_muxps_plot()
        self.key_handler: int = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.close_window,
            )
        signals.emit(Signal.BLOCK_KEYBINDINGS, window=self.window, window_object=self)

    def close_window(self):
        if dpg.does_item_exist(self.window):
            dpg.delete_item(self.window)
        if dpg.does_item_exist(self.key_handler):
            dpg.delete_item(self.key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    def generate_data(
        self,
        noise: Optional[ndarray] = None,
    ) -> Tuple[DataSet, DataSet, DataSet, ndarray, ndarray, ndarray, DRTResult]:
        circuit: Circuit = pyimpspec.parse_cdc("R{R=100}(C{C=1e-6}[R{R=500}W{Y=4e-4}])")
        data: DataSet = pyimpspec.simulate_spectrum(
            circuit,
            logspace(
                0,
                5,
                num=23,
            ),
        )
        Z: complex
        sd: float = 0.01
        if noise is None:
            noise = array(
                list(
                    map(
                        lambda Z: complex(
                            abs(Z) * normal(0, sd, 1),
                            abs(Z) * normal(0, sd, 1),
                        ),
                        data.get_impedance(),
                    )
                )
            )
        data.subtract_impedance(-noise)
        sim_data: DataSet = pyimpspec.simulate_spectrum(
            circuit,
            data.get_frequency(),
        )
        smooth_data: DataSet = pyimpspec.simulate_spectrum(
            circuit,
            logspace(
                0,
                5,
                num=5 * STATE.config.num_per_decade_in_simulated_lines + 1,
            ),
        )
        real_residual: ndarray = (
            (data.get_impedance().real - sim_data.get_impedance().real)
            / abs(data.get_impedance())
            * 100
        )
        imag_residual: ndarray = (
            (data.get_impedance().imag - sim_data.get_impedance().imag)
            / abs(data.get_impedance())
            * 100
        )
        f: ndarray = smooth_data.get_frequency()
        tau: ndarray = 1 / f
        gamma: ndarray = array(
            list(
                map(
                    lambda _, a=1.1, b=0.0001, c=0.00001: a
                    * exp(-((_ - b) ** 2) / (2 * c**2)),
                    tau,
                )
            )
        )
        mean_gamma: ndarray = array(
            list(
                map(
                    lambda _, a=0.9, b=0.012, c=0.001: a
                    * exp(-((_ - b) ** 2) / (2 * c**2)),
                    tau,
                )
            )
        )
        lower_bound: ndarray = array(
            list(
                map(
                    lambda _, a=0.4, b=0.012, c=0.001: a
                    * exp(-((_ - b) ** 2) / (2 * c**2)),
                    tau,
                )
            )
        )
        upper_bound: ndarray = array(
            list(
                map(
                    lambda _, a=1.4, b=0.012, c=0.002: a
                    * exp(-((_ - b) ** 2) / (2 * c**2)),
                    tau,
                )
            )
        )
        imaginary_gamma: ndarray = array(
            list(
                map(
                    lambda _, a=1.0, b=0.001, c=0.0001: a
                    * exp(-((_ - b) ** 2) / (2 * c**2)),
                    tau,
                )
            )
        )
        drt: DRTResult = DRTResult(
            uuid4().hex,
            time(),
            tau,
            gamma,
            f,
            sim_data.get_impedance(),
            real_residual,
            imag_residual,
            mean_gamma,
            lower_bound,
            upper_bound,
            imaginary_gamma,
            {},
            0.0,
            0.0,
            {},
            DRTSettings(
                method=DRTMethod.BHT,
                mode=DRTMode.COMPLEX,
                lambda_value=0.0,
                rbf_type=RBFType.GAUSSIAN,
                derivative_order=1,
                rbf_shape=RBFShape.FWHM,
                shape_coeff=0.5,
                inductance=False,
                credible_intervals=True,
                num_samples=2000,
                num_attempts=50,
                maximum_symmetry=0.5,
                circuit=None,
                W=0.15,
                num_per_decade=100,
            ),
        )
        return (
            data,
            sim_data,
            smooth_data,
            real_residual,
            imag_residual,
            noise,
            drt,
        )

    def update_simulated_num_per_decade(self, sender: int, value: int):
        STATE.config.num_per_decade_in_simulated_lines = value
        (
            self.data,
            self.sim_data,
            self.smooth_data,
            self.real_residual,
            self.imag_residual,
            _,
            _,
        ) = self.generate_data(self.noise)
        self.update_bode_plot()
        self.update_drt_plot()
        self.update_impedance_plot()
        self.update_nyquist_plot()
        self.update_residuals_plot()
        self.update_muxps_plot()

    def update_bode_plot(self, adjust_limits: bool = False):
        self.bode_plot.clear()
        freq, mag, phase = self.data.get_bode_data()
        self.bode_plot.plot(
            frequency=freq,
            magnitude=mag,
            phase=phase,
            labels=(
                "|Z| (d)",
                "phi (d)",
            ),
            themes=(
                themes.bode.magnitude_data,
                themes.bode.phase_data,
            ),
        )
        freq, mag, phase = self.smooth_data.get_bode_data()
        self.bode_plot.plot(
            frequency=freq,
            magnitude=mag,
            phase=phase,
            labels=(
                "|Z| (f)",
                "phi (f)",
            ),
            line=True,
            themes=(
                themes.bode.magnitude_simulation,
                themes.bode.phase_simulation,
            ),
        )
        freq, mag, phase = self.sim_data.get_bode_data()
        self.bode_plot.plot(
            frequency=freq,
            magnitude=mag,
            phase=phase,
            labels=(
                "|Z| (f)",
                "phi (f)",
            ),
            line=False,
            show_labels=False,
            themes=(
                themes.bode.magnitude_simulation,
                themes.bode.phase_simulation,
            ),
        )
        if adjust_limits:
            self.bode_plot.adjust_limits()

    def update_bode_color(self, sender: int, _, theme: int):
        assert type(sender) is int
        assert type(theme) is int
        color: List[float] = dpg.get_value(sender)
        themes.update_plot_series_theme_color(theme, color)
        STATE.config.colors[
            {
                themes.bode.magnitude_data: "bode_magnitude_data",
                themes.bode.phase_data: "bode_phase_data",
                themes.bode.magnitude_simulation: "bode_magnitude_simulation",
                themes.bode.phase_simulation: "bode_phase_simulation",
            }[theme]
        ] = color

    def update_bode_marker(self, sender: int, label: str, theme: int):
        assert type(sender) is int
        assert type(label) is str
        assert type(theme) is int
        marker: int = PLOT_MARKERS[label]
        themes.update_plot_series_theme_marker(theme, marker)
        STATE.config.markers[
            {
                themes.bode.magnitude_data: "bode_magnitude_data",
                themes.bode.phase_data: "bode_phase_data",
                themes.bode.magnitude_simulation: "bode_magnitude_simulation",
                themes.bode.phase_simulation: "bode_phase_simulation",
            }[theme]
        ] = marker

    def reset_bode_plot(self):
        dpg.set_value(
            self.bode_data_mag_color,
            (
                51,
                187,
                238,
                190,
            ),
        )
        dpg.set_value(
            self.bode_data_phase_color,
            (
                238,
                119,
                51,
                190,
            ),
        )
        dpg.set_value(
            self.bode_sim_mag_color,
            (
                238,
                51,
                119,
                190,
            ),
        )
        dpg.set_value(
            self.bode_sim_phase_color,
            (
                0,
                153,
                136,
                190,
            ),
        )
        self.update_bode_color(
            self.bode_data_mag_color, None, themes.bode.magnitude_data
        )
        self.update_bode_color(self.bode_data_phase_color, None, themes.bode.phase_data)
        self.update_bode_color(
            self.bode_sim_mag_color, None, themes.bode.magnitude_simulation
        )
        self.update_bode_color(
            self.bode_sim_phase_color, None, themes.bode.phase_simulation
        )
        dpg.set_value(self.bode_data_mag_marker, "Circle")
        dpg.set_value(self.bode_data_phase_marker, "Square")
        dpg.set_value(self.bode_sim_mag_marker, "Cross")
        dpg.set_value(self.bode_sim_phase_marker, "Plus")
        self.update_bode_marker(
            self.bode_data_mag_marker,
            dpg.get_value(self.bode_data_mag_marker),
            themes.bode.magnitude_data,
        )
        self.update_bode_marker(
            self.bode_data_phase_marker,
            dpg.get_value(self.bode_data_phase_marker),
            themes.bode.phase_data,
        )
        self.update_bode_marker(
            self.bode_sim_mag_marker,
            dpg.get_value(self.bode_sim_mag_marker),
            themes.bode.magnitude_simulation,
        )
        self.update_bode_marker(
            self.bode_sim_phase_marker,
            dpg.get_value(self.bode_sim_phase_marker),
            themes.bode.phase_simulation,
        )

    def update_drt_plot(self, adjust_limits: bool = False):
        self.drt_plot.clear()
        tau: ndarray
        gamma: ndarray
        tau, gamma = self.drt.get_drt_data()
        self.drt_plot.plot(
            tau=tau,
            gamma=gamma,
            label="gamma/real",
            theme=themes.drt.real_gamma,
        )
        tau, gamma = self.drt.get_drt_data(imaginary=True)
        self.drt_plot.plot(
            tau=tau,
            gamma=gamma,
            label="imag.",
            theme=themes.drt.imaginary_gamma,
        )
        lower: ndarray
        upper: ndarray
        tau, gamma, lower, upper = self.drt.get_drt_credible_intervals()
        self.drt_plot.plot(
            tau=tau,
            gamma=gamma,
            label="mean",
            theme=themes.drt.mean_gamma,
        )
        self.drt_plot.plot(
            tau=tau,
            lower=lower,
            upper=upper,
            label="3-sigma CI",
            theme=themes.drt.credible_intervals,
        )
        if adjust_limits:
            self.drt_plot.adjust_limits()

    def update_drt_color(self, sender: int, _, theme: int):
        assert type(sender) is int
        assert type(theme) is int
        color: List[float] = dpg.get_value(sender)
        themes.update_plot_series_theme_color(theme, color)
        STATE.config.colors[
            {
                themes.drt.real_gamma: "drt_real_gamma",
                themes.drt.imaginary_gamma: "drt_imaginary_gamma",
                themes.drt.mean_gamma: "drt_mean_gamma",
                themes.drt.credible_intervals: "drt_credible_intervals",
            }[theme]
        ] = color

    def reset_drt_plot(self):
        dpg.set_value(
            self.drt_real_gamma_color,
            (
                51.0,
                187.0,
                238.0,
                190.0,
            ),
        )
        dpg.set_value(
            self.drt_imaginary_gamma_color,
            (
                238.0,
                119.0,
                51.0,
                190.0,
            ),
        )
        dpg.set_value(
            self.drt_mean_gamma_color,
            (
                238.0,
                119.0,
                51.0,
                190.0,
            ),
        )
        dpg.set_value(
            self.drt_credible_intervals_color,
            (
                238.0,
                119.0,
                51.0,
                48.0,
            ),
        )
        self.update_drt_color(
            self.drt_real_gamma_color,
            None,
            themes.drt.real_gamma,
        )
        self.update_drt_color(
            self.drt_imaginary_gamma_color,
            None,
            themes.drt.imaginary_gamma,
        )
        self.update_drt_color(
            self.drt_mean_gamma_color,
            None,
            themes.drt.mean_gamma,
        )
        self.update_drt_color(
            self.drt_credible_intervals_color,
            None,
            themes.drt.credible_intervals,
        )

    def update_impedance_plot(self, adjust_limits: bool = False):
        self.impedance_plot.clear()
        f: ndarray = self.data.get_frequency()
        Z: ndarray = self.data.get_impedance()
        self.impedance_plot.plot(
            frequency=f,
            real=Z.real,
            imaginary=-Z.imag,
            labels=(
                "Z' (d)",
                'Z" (d)',
            ),
            themes=(
                themes.impedance.real_data,
                themes.impedance.imaginary_data,
            ),
        )
        f = self.sim_data.get_frequency()
        Z = self.sim_data.get_impedance()
        self.impedance_plot.plot(
            frequency=f,
            real=Z.real,
            imaginary=-Z.imag,
            labels=(
                "Z' (f)",
                'Z" (f)',
            ),
            fit=True,
            line=False,
            themes=(
                themes.impedance.real_simulation,
                themes.impedance.imaginary_simulation,
            ),
        )
        f = self.smooth_data.get_frequency()
        Z = self.smooth_data.get_impedance()
        self.impedance_plot.plot(
            frequency=f,
            real=Z.real,
            imaginary=-Z.imag,
            labels=(
                "Z' (f)",
                'Z" (f)',
            ),
            fit=True,
            line=True,
            themes=(
                themes.impedance.real_simulation,
                themes.impedance.imaginary_simulation,
            ),
            show_labels=False,
        )
        if adjust_limits:
            self.impedance_plot.adjust_limits()

    def update_impedance_color(self, sender: int, _, theme: int):
        assert type(sender) is int
        assert type(theme) is int
        color: List[float] = dpg.get_value(sender)
        themes.update_plot_series_theme_color(theme, color)
        STATE.config.colors[
            {
                themes.impedance.real_data: "impedance_real_data",
                themes.impedance.real_simulation: "impedance_real_simulation",
                themes.impedance.imaginary_data: "impedance_imaginary_data",
                themes.impedance.imaginary_simulation: "impedance_imaginary_simulation",
            }[theme]
        ] = color

    def update_impedance_marker(self, sender: int, label: str, theme: int):
        assert type(sender) is int
        assert type(label) is str
        assert type(theme) is int
        marker: int = PLOT_MARKERS[label]
        themes.update_plot_series_theme_marker(theme, marker)
        STATE.config.markers[
            {
                themes.impedance.real_data: "impedance_real_data",
                themes.impedance.real_simulation: "impedance_real_simulation",
                themes.impedance.imaginary_data: "impedance_imaginary_data",
                themes.impedance.imaginary_simulation: "impedance_imaginary_simulation",
            }[theme]
        ] = marker

    def reset_impedance_plot(self):
        dpg.set_value(
            self.impedance_real_data_color,
            (
                51.0,
                187.0,
                238.0,
                190.0,
            ),
        )
        dpg.set_value(
            self.impedance_real_simulation_color,
            (
                238.0,
                51.0,
                119.0,
                190.0,
            ),
        )
        dpg.set_value(
            self.impedance_imaginary_data_color,
            (
                238.0,
                119.0,
                51.0,
                190.0,
            ),
        )
        dpg.set_value(
            self.impedance_imaginary_simulation_color,
            (
                0.0,
                153.0,
                136.0,
                190.0,
            ),
        )
        self.update_impedance_color(
            self.impedance_real_data_color,
            None,
            themes.impedance.real_data,
        )
        self.update_impedance_color(
            self.impedance_real_simulation_color,
            None,
            themes.impedance.real_simulation,
        )
        self.update_impedance_color(
            self.impedance_imaginary_data_color,
            None,
            themes.impedance.imaginary_data,
        )
        self.update_impedance_color(
            self.impedance_imaginary_simulation_color,
            None,
            themes.impedance.imaginary_simulation,
        )
        dpg.set_value(self.impedance_real_data_marker, "Circle")
        dpg.set_value(self.impedance_real_simulation_marker, "Cross")
        dpg.set_value(self.impedance_imaginary_data_marker, "Square")
        dpg.set_value(self.impedance_imaginary_simulation_marker, "Plus")
        self.update_impedance_marker(
            self.impedance_real_data_marker,
            dpg.get_value(self.impedance_real_data_marker),
            themes.impedance.real_data,
        )
        self.update_impedance_marker(
            self.impedance_real_simulation_marker,
            dpg.get_value(self.impedance_real_simulation_marker),
            themes.impedance.real_simulation,
        )
        self.update_impedance_marker(
            self.impedance_imaginary_data_marker,
            dpg.get_value(self.impedance_imaginary_data_marker),
            themes.impedance.imaginary_data,
        )
        self.update_impedance_marker(
            self.impedance_imaginary_simulation_marker,
            dpg.get_value(self.impedance_imaginary_simulation_marker),
            themes.impedance.imaginary_simulation,
        )

    def update_nyquist_plot(self, adjust_limits: bool = False):
        self.nyquist_plot.clear()
        real, imag = self.data.get_nyquist_data()
        self.nyquist_plot.plot(
            real=real,
            imaginary=imag,
            label="Data",
            theme=themes.nyquist.data,
        )
        real, imag = self.smooth_data.get_nyquist_data()
        self.nyquist_plot.plot(
            real=real,
            imaginary=imag,
            label="Fit",
            line=True,
            theme=themes.nyquist.simulation,
        )
        real, imag = self.sim_data.get_nyquist_data()
        self.nyquist_plot.plot(
            real=real,
            imaginary=imag,
            label="Fit",
            show_label=False,
            theme=themes.nyquist.simulation,
        )
        if adjust_limits:
            self.nyquist_plot.adjust_limits()

    def update_nyquist_color(self, sender: int, _, theme: int):
        assert type(sender) is int
        assert type(theme) is int
        color: List[float] = dpg.get_value(sender)
        themes.update_plot_series_theme_color(theme, color)
        STATE.config.colors[
            {
                themes.nyquist.data: "nyquist_data",
                themes.nyquist.simulation: "nyquist_simulation",
            }[theme]
        ] = color

    def update_nyquist_marker(self, sender: int, label: str, theme: int):
        assert type(sender) is int
        assert type(label) is str
        assert type(theme) is int
        marker: int = PLOT_MARKERS[label]
        themes.update_plot_series_theme_marker(theme, marker)
        STATE.config.markers[
            {
                themes.nyquist.data: "nyquist_data",
                themes.nyquist.simulation: "nyquist_simulation",
            }[theme]
        ] = marker

    def reset_nyquist_plot(self):
        dpg.set_value(
            self.nyquist_data_color,
            (
                51,
                187,
                238,
                190,
            ),
        )
        dpg.set_value(
            self.nyquist_sim_color,
            (
                238,
                51,
                119,
                190,
            ),
        )
        self.update_nyquist_color(self.nyquist_data_color, None, themes.nyquist.data)
        self.update_nyquist_color(
            self.nyquist_sim_color, None, themes.nyquist.simulation
        )
        dpg.set_value(self.nyquist_data_marker, "Circle")
        dpg.set_value(self.nyquist_sim_marker, "Cross")
        self.update_nyquist_marker(
            self.nyquist_data_marker,
            dpg.get_value(self.nyquist_data_marker),
            themes.nyquist.data,
        )
        self.update_nyquist_marker(
            self.nyquist_sim_marker,
            dpg.get_value(self.nyquist_sim_marker),
            themes.nyquist.simulation,
        )

    def update_residuals_plot(self, adjust_limits: bool = False):
        self.residuals_plot.clear()
        self.residuals_plot.plot(
            frequency=self.data.get_frequency(),
            real=self.real_residual,
            imaginary=self.imag_residual,
        )
        if adjust_limits:
            self.residuals_plot.adjust_limits()

    def update_residuals_color(self, sender: int, _, theme: int):
        assert type(sender) is int
        assert type(theme) is int
        color: List[float] = dpg.get_value(sender)
        themes.update_plot_series_theme_color(theme, color)
        STATE.config.colors[
            {
                themes.residuals.real: "residuals_real",
                themes.residuals.imaginary: "residuals_imaginary",
            }[theme]
        ] = color

    def update_residuals_marker(self, sender: int, label: str, theme: int):
        assert type(sender) is int
        assert type(label) is str
        assert type(theme) is int
        marker: int = PLOT_MARKERS[label]
        themes.update_plot_series_theme_marker(theme, marker)
        STATE.config.markers[
            {
                themes.residuals.real: "residuals_real",
                themes.residuals.imaginary: "residuals_imaginary",
            }[theme]
        ] = marker

    def reset_residuals_plot(self):
        dpg.set_value(
            self.residuals_real_color,
            (
                238,
                51,
                119,
                190,
            ),
        )
        dpg.set_value(
            self.residuals_imag_color,
            (
                0,
                153,
                136,
                190,
            ),
        )
        self.update_residuals_color(
            self.residuals_real_color, None, themes.residuals.real
        )
        self.update_residuals_color(
            self.residuals_imag_color, None, themes.residuals.imaginary
        )
        dpg.set_value(self.residuals_real_marker, "Circle")
        dpg.set_value(self.residuals_imag_marker, "Square")
        self.update_residuals_marker(
            self.residuals_real_marker,
            dpg.get_value(self.residuals_real_marker),
            themes.residuals.real,
        )
        self.update_residuals_marker(
            self.residuals_imag_marker,
            dpg.get_value(self.residuals_imag_marker),
            themes.residuals.imaginary,
        )

    def update_muxps_plot(self):
        self.muxps_plot.clear()
        self.muxps_plot.plot(
            num_RCs=array([1, 2, 3, 4, 5, 6]),
            mu=array([1.0, 0.8, 0.6, 0.4, 0.2, 0.0]),
            Xps=array([0, -1, -1.5, -1.8, -2, -1.4]),
            mu_criterion=0.7,
            num_RC=3,
        )
        self.muxps_plot.adjust_limits()

    def update_muxps_color(self, sender: int, _, theme: int):
        assert type(sender) is int
        assert type(theme) is int
        color: List[float] = dpg.get_value(sender)
        themes.update_plot_series_theme_color(theme, color)
        STATE.config.colors[
            {
                themes.mu_Xps.mu_criterion: "mu_Xps_mu_criterion",
                themes.mu_Xps.mu: "mu_Xps_mu",
                themes.mu_Xps.mu_highlight: "mu_Xps_mu_highlight",
                themes.mu_Xps.Xps: "mu_Xps_Xps",
                themes.mu_Xps.Xps_highlight: "mu_Xps_Xps_highlight",
            }[theme]
        ] = color

    def update_muxps_marker(self, sender: int, label: str, theme: int):
        assert type(sender) is int
        assert type(label) is str
        assert type(theme) is int
        marker: int = PLOT_MARKERS[label]
        themes.update_plot_series_theme_marker(theme, marker)
        STATE.config.markers[
            {
                themes.mu_Xps.mu: "mu_Xps_mu",
                themes.mu_Xps.Xps: "mu_Xps_Xps",
            }[theme]
        ] = marker
        theme = {
            themes.mu_Xps.mu: themes.mu_Xps.mu_highlight,
            themes.mu_Xps.Xps: themes.mu_Xps.Xps_highlight,
        }[theme]
        themes.update_plot_series_theme_marker(theme, marker)
        STATE.config.markers[
            {
                themes.mu_Xps.mu_highlight: "mu_Xps_mu_highlight",
                themes.mu_Xps.Xps_highlight: "mu_Xps_Xps_highlight",
            }[theme]
        ] = marker

    def reset_muxps_plot(self):
        dpg.set_value(
            self.muxps_mu_criterion_color,
            (
                255,
                255,
                255,
                128,
            ),
        )
        dpg.set_value(
            self.muxps_mu_color,
            (
                238,
                51,
                119,
                190,
            ),
        )
        dpg.set_value(
            self.muxps_mu_highlight_color,
            (
                51,
                187,
                238,
                190,
            ),
        )
        dpg.set_value(
            self.muxps_xps_color,
            (
                0,
                153,
                136,
                190,
            ),
        )
        dpg.set_value(
            self.muxps_xps_highlight_color,
            (
                238,
                119,
                51,
                190,
            ),
        )
        self.update_muxps_color(
            self.muxps_mu_criterion_color, None, themes.mu_Xps.mu_criterion
        )
        self.update_muxps_color(self.muxps_mu_color, None, themes.mu_Xps.mu)
        self.update_muxps_color(
            self.muxps_mu_highlight_color, None, themes.mu_Xps.mu_highlight
        )
        self.update_muxps_color(self.muxps_xps_color, None, themes.mu_Xps.Xps)
        self.update_muxps_color(
            self.muxps_xps_highlight_color,
            None,
            themes.mu_Xps.Xps_highlight,
        )
        dpg.set_value(self.muxps_mu_marker, "Circle")
        dpg.set_value(self.muxps_xps_marker, "Square")
        self.update_muxps_marker(
            self.muxps_mu_marker,
            dpg.get_value(self.muxps_mu_marker),
            themes.mu_Xps.mu,
        )
        self.update_muxps_marker(
            self.muxps_xps_marker,
            dpg.get_value(self.muxps_xps_marker),
            themes.mu_Xps.Xps,
        )

# TODO: Settings for circuit diagrams
# - WE and CE+RE labels



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import Callable, Dict, List, Optional
from numpy import allclose, array, log10 as log, ndarray
from pyimpspec.analysis.kramers_kronig import _calculate_residuals
import dearpygui.dearpygui as dpg
from deareis.signals import Signal
import deareis.signals as signals
from deareis.enums import (
    Context,
    CNLSMethod,
    TestMode,
    Test,
    label_to_cnls_method,
    label_to_test_mode,
    label_to_test,
    cnls_method_to_label,
    test_mode_to_label,
    test_to_label,
)
from deareis.data import TestResult, TestSettings, DataSet
from deareis.gui.plots import Bode, Nyquist, Residuals
import deareis.tooltips as tooltips
from deareis.tooltips import attach_tooltip, update_tooltip
import deareis.themes as themes


class SettingsMenu:
    def __init__(
        self, default_settings: TestSettings, label_pad: int, limited: bool = False
    ):
        with dpg.group(horizontal=True):
            dpg.add_text("Test".rjust(label_pad))
            attach_tooltip(tooltips.kramers_kronig.test)
            self.test_combo: int = dpg.generate_uuid()
            dpg.add_combo(
                default_value=test_to_label[default_settings.test],
                items=list(label_to_test.keys()),
                callback=lambda s, a, u: self.update_settings(),
                width=-1,
                tag=self.test_combo,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Mode".rjust(label_pad))
            attach_tooltip(tooltips.kramers_kronig.mode)
            self.mode_combo: int = dpg.generate_uuid()
            dpg.add_combo(
                default_value=test_mode_to_label[default_settings.mode],
                items=list(label_to_test_mode.keys()),
                callback=lambda s, a, u: self.update_settings(),
                width=-1,
                tag=self.mode_combo,
            )
        with dpg.group(horizontal=True, show=not limited):
            self.num_RC_label: int = dpg.generate_uuid()
            num_RC_labels: List[str] = [
                "Max. num. RC elements".rjust(label_pad),
                "Number of RC elements".rjust(label_pad),
            ]
            dpg.add_text(
                num_RC_labels[0],
                user_data=num_RC_labels,
                tag=self.num_RC_label,
            )
            attach_tooltip(tooltips.kramers_kronig.max_num_RC)
            self.num_RC_slider: int = dpg.generate_uuid()
            dpg.add_slider_int(
                clamped=True,
                width=-1,
                tag=self.num_RC_slider,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Add capacitor in series".rjust(label_pad))
            attach_tooltip(tooltips.kramers_kronig.add_capacitance)
            self.add_cap_checkbox: int = dpg.generate_uuid()
            dpg.add_checkbox(
                default_value=default_settings.add_capacitance,
                tag=self.add_cap_checkbox,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Add inductor in series".rjust(label_pad))
            attach_tooltip(tooltips.kramers_kronig.add_inductance)
            self.add_ind_checkbox: int = dpg.generate_uuid()
            dpg.add_checkbox(
                default_value=default_settings.add_inductance,
                tag=self.add_ind_checkbox,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("µ-criterion".rjust(label_pad))
            attach_tooltip(tooltips.kramers_kronig.mu_criterion)
            self.mu_crit_slider: int = dpg.generate_uuid()
            dpg.add_slider_float(
                default_value=default_settings.mu_criterion,
                min_value=0.01,
                max_value=0.99,
                clamped=True,
                format="%.2f",
                width=-1,
                tag=self.mu_crit_slider,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Fitting method".rjust(label_pad))
            attach_tooltip(tooltips.kramers_kronig.method)
            self.method_combo: int = dpg.generate_uuid()
            dpg.add_combo(
                default_value=cnls_method_to_label.get(
                    default_settings.method,
                    list(label_to_cnls_method.keys())[0],
                ),
                items=list(label_to_cnls_method.keys()),
                width=-1,
                tag=self.method_combo,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Max. num. of func. eval.".rjust(label_pad))
            attach_tooltip(tooltips.kramers_kronig.nfev)
            self.max_nfev_input: int = dpg.generate_uuid()
            dpg.add_input_int(
                default_value=default_settings.max_nfev,
                min_value=0,
                min_clamped=True,
                step=0,
                on_enter=True,
                width=-1,
                tag=self.max_nfev_input,
            )
        self.update_settings()

    def get_settings(self) -> TestSettings:
        return TestSettings(
            test=label_to_test.get(dpg.get_value(self.test_combo), Test.COMPLEX),
            mode=label_to_test_mode.get(
                dpg.get_value(self.mode_combo), TestMode.EXPLORATORY
            ),
            num_RC=dpg.get_value(self.num_RC_slider),
            mu_criterion=dpg.get_value(self.mu_crit_slider),
            add_capacitance=dpg.get_value(self.add_cap_checkbox),
            add_inductance=dpg.get_value(self.add_ind_checkbox),
            method=label_to_cnls_method.get(
                dpg.get_value(self.method_combo), CNLSMethod.LEASTSQ
            ),
            max_nfev=dpg.get_value(self.max_nfev_input),
        )

    def set_settings(self, settings: TestSettings):
        assert type(settings) is TestSettings, settings
        dpg.set_value(self.test_combo, test_to_label.get(settings.test))
        dpg.set_value(self.mode_combo, test_mode_to_label.get(settings.mode))
        num_RC: int = settings.num_RC
        item_configuration: dict = dpg.get_item_configuration(self.num_RC_slider)
        if num_RC >= 2:
            if num_RC > item_configuration["max_value"]:
                num_RC = item_configuration["max_value"]
            dpg.set_value(self.num_RC_slider, num_RC)
        dpg.set_value(self.add_cap_checkbox, settings.add_capacitance)
        dpg.set_value(self.add_ind_checkbox, settings.add_inductance)
        dpg.set_value(self.mu_crit_slider, settings.mu_criterion)
        dpg.set_value(self.method_combo, cnls_method_to_label.get(settings.method))
        dpg.set_value(self.max_nfev_input, settings.max_nfev)
        self.update_settings()

    def update_settings(self, settings: Optional[TestSettings] = None):
        if settings is None:
            settings = self.get_settings()
        if settings.test == Test.CNLS:
            dpg.enable_item(self.add_ind_checkbox)
            dpg.enable_item(self.method_combo)
            dpg.enable_item(self.max_nfev_input)
        else:
            dpg.disable_item(self.add_ind_checkbox)
            dpg.set_value(self.add_ind_checkbox, True)
            dpg.disable_item(self.method_combo)
            dpg.disable_item(self.max_nfev_input)
        if settings.mode == TestMode.MANUAL:
            dpg.disable_item(self.mu_crit_slider)
            dpg.set_value(
                self.num_RC_label, dpg.get_item_user_data(self.num_RC_label)[1]
            )
        else:
            dpg.enable_item(self.mu_crit_slider)
            dpg.set_value(
                self.num_RC_label, dpg.get_item_user_data(self.num_RC_label)[0]
            )

    def update_num_RC_slider(self, data: Optional[DataSet]):
        min_num_RC: int = 0 if data is None else 2
        max_num_RC: int = 9999 if data is None else data.get_num_points()
        num_RC: int = dpg.get_value(self.num_RC_slider)
        if num_RC > max_num_RC:
            num_RC = max_num_RC
        elif num_RC < min_num_RC:
            num_RC = max_num_RC
        dpg.configure_item(
            self.num_RC_slider,
            default_value=num_RC,
            min_value=min_num_RC,
            max_value=max_num_RC,
        )

    def get_num_RC_labels(self) -> List[str]:
        return dpg.get_item_user_data(self.num_RC_label)

    def has_active_input(self) -> bool:
        return dpg.is_item_active(self.max_nfev_input)


class StatisticsTable:
    def __init__(self):
        label_pad: int = 23
        self._header: int = dpg.generate_uuid()
        with dpg.collapsing_header(label=" Statistics", leaf=True, tag=self._header):
            self._table: int = dpg.generate_uuid()
            with dpg.table(
                borders_outerV=True,
                borders_outerH=True,
                borders_innerV=True,
                borders_innerH=True,
                scrollY=True,
                freeze_rows=1,
                height=18 + 23 * 3,
                tag=self._table,
            ):
                dpg.add_table_column(
                    label="Label".rjust(label_pad),
                    width_fixed=True,
                )
                dpg.add_table_column(
                    label="Value",
                    width_fixed=True,
                )
                label: str
                tooltip: str
                for (label, tooltip) in [
                    ("log X² (pseudo)", tooltips.kramers_kronig.pseudo_chisqr),
                    ("µ", tooltips.kramers_kronig.mu),
                    ("Number of RC elements", tooltips.kramers_kronig.num_RC),
                ]:
                    with dpg.table_row(parent=self._table):
                        dpg.add_text(label.rjust(label_pad))
                        attach_tooltip(tooltip)
                        tooltip_tag: int = dpg.generate_uuid()
                        dpg.add_text("", user_data=tooltip_tag)
                        attach_tooltip("", tag=tooltip_tag)

    def clear(self, hide: bool):
        if hide:
            dpg.hide_item(self._header)
        row: int
        for row in dpg.get_item_children(self._table, slot=1):
            tag: int = dpg.get_item_children(row, slot=1)[2]
            dpg.set_value(tag, "")
            dpg.hide_item(dpg.get_item_parent(dpg.get_item_user_data(tag)))

    def populate(self, test: TestResult):
        dpg.show_item(self._header)
        cells: List[int] = []
        row: int
        for row in dpg.get_item_children(self._table, slot=1):
            cells.append(dpg.get_item_children(row, slot=1)[2])
        assert len(cells) == 3, cells
        tag: int
        value: str
        for (tag, value) in [
            (
                cells[0],
                f"{log(test.pseudo_chisqr):.3f}",
            ),
            (
                cells[1],
                f"{test.mu:.3f}",
            ),
            (
                cells[2],
                f"{test.num_RC}",
            ),
        ]:
            dpg.set_value(tag, value)
            update_tooltip(dpg.get_item_user_data(tag), value)
            dpg.show_item(dpg.get_item_parent(dpg.get_item_user_data(tag)))


class SettingsTable:
    def __init__(self):
        label_pad: int = 23
        self._header: int = dpg.generate_uuid()
        with dpg.collapsing_header(label=" Settings", leaf=True, tag=self._header):
            self._table: int = dpg.generate_uuid()
            with dpg.table(
                borders_outerV=True,
                borders_outerH=True,
                borders_innerV=True,
                borders_innerH=True,
                scrollY=True,
                freeze_rows=1,
                height=18 + 23 * 8,
                tag=self._table,
            ):
                dpg.add_table_column(
                    label="Label".rjust(label_pad),
                    width_fixed=True,
                )
                dpg.add_table_column(
                    label="Value",
                    width_fixed=True,
                )
                label: str
                for label in [
                    "Test",
                    "Mode",
                    "Number of RC elements",
                    "Add capacitor in series",
                    "Add inductor in series",
                    "µ-criterion",
                    "Fitting method",
                    "Max. num. func. eval.",
                ]:
                    with dpg.table_row(parent=self._table):
                        dpg.add_text(label.rjust(label_pad))
                        tooltip_tag: int = dpg.generate_uuid()
                        dpg.add_text("", user_data=tooltip_tag)
                        attach_tooltip("", tag=tooltip_tag)
            with dpg.group(horizontal=True):
                self._apply_settings_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Apply settings",
                    callback=lambda s, a, u: signals.emit(
                        Signal.APPLY_TEST_SETTINGS,
                        **u,
                    ),
                    tag=self._apply_settings_button,
                )
                attach_tooltip(tooltips.general.apply_settings)
                self._apply_mask_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Apply mask",
                    callback=lambda s, a, u: signals.emit(
                        Signal.APPLY_DATA_SET_MASK,
                        **u,
                    ),
                    tag=self._apply_mask_button,
                )
                attach_tooltip(tooltips.general.apply_mask)

    def clear(self, hide: bool):
        if hide:
            dpg.hide_item(self._header)
        row: int
        for row in dpg.get_item_children(self._table, slot=1):
            tag: int = dpg.get_item_children(row, slot=1)[1]
            dpg.set_value(tag, "")
            dpg.hide_item(dpg.get_item_parent(dpg.get_item_user_data(tag)))

    def populate(self, test: TestResult, data: DataSet, num_RC_labels: List[str]):
        dpg.show_item(self._header)
        rows: List[int] = []
        cells: List[List[int]] = []
        row: int
        for row in dpg.get_item_children(self._table, slot=1):
            rows.append(row)
            cells.append(dpg.get_item_children(row, slot=1))
        assert len(rows) == len(cells) == 8, (
            rows,
            cells,
        )
        if test.settings.mode == TestMode.MANUAL:
            dpg.set_value(
                cells[1][0],
                num_RC_labels[1],
            )
        else:
            dpg.set_value(
                cells[1][0],
                num_RC_labels[0],
            )
        num_rows: int = 0
        tag: int
        value: str
        visible: bool
        for (row, tag, value, visible) in [
            (
                rows[0],
                cells[0][1],
                test_to_label.get(test.settings.test, ""),
                True,
            ),
            (
                rows[1],
                cells[1][1],
                test_mode_to_label.get(test.settings.mode, ""),
                True,
            ),
            (
                rows[2],
                cells[2][1],
                f"{test.settings.num_RC}",
                True,
            ),
            (
                rows[3],
                cells[3][1],
                "True" if test.settings.add_capacitance else "False",
                test.settings.add_capacitance,
            ),
            (
                rows[4],
                cells[4][1],
                "True" if test.settings.add_inductance else "False",
                test.settings.add_inductance,
            ),
            (
                rows[5],
                cells[5][1],
                f"{test.settings.mu_criterion:.2f}",
                test.settings.mode != TestMode.MANUAL,
            ),
            (
                rows[6],
                cells[6][1],
                cnls_method_to_label.get(test.settings.method, ""),
                test.settings.test == Test.CNLS,
            ),
            (
                rows[7],
                cells[7][1],
                f"{test.settings.max_nfev}",
                test.settings.test == Test.CNLS,
            ),
        ]:
            dpg.set_value(tag, value)
            update_tooltip(dpg.get_item_user_data(tag), value)
            dpg.show_item(dpg.get_item_parent(dpg.get_item_user_data(tag)))
            if visible:
                dpg.show_item(row)
                num_rows += 1
            else:
                dpg.hide_item(row)
        dpg.set_item_height(self._table, 18 + 23 * max(1, num_rows))
        dpg.set_item_user_data(
            self._apply_settings_button,
            {
                "settings": test.settings,
            },
        )
        dpg.set_item_user_data(
            self._apply_mask_button,
            {
                "data": data,
                "mask": test.mask,
                "test": test,
            },
        )


class DataSetsCombo:
    def __init__(self, label: str, width: int):
        self.labels: List[str] = []
        dpg.add_text(label)
        self.tag: int = dpg.generate_uuid()
        dpg.add_combo(
            callback=lambda s, a, u: signals.emit(
                Signal.SELECT_DATA_SET,
                data=u.get(a),
            ),
            user_data={},
            width=width,
            tag=self.tag,
        )

    def populate(self, labels: List[str], lookup: Dict[str, DataSet]):
        self.labels.clear()
        self.labels.extend(labels)
        label: str = dpg.get_value(self.tag) or ""
        if labels and label not in labels:
            label = labels[0]
        dpg.configure_item(
            self.tag,
            default_value=label,
            items=labels,
            user_data=lookup,
        )

    def get(self) -> Optional[DataSet]:
        return dpg.get_item_user_data(self.tag).get(dpg.get_value(self.tag))

    def set(self, label: str):
        assert type(label) is str, label
        assert label in self.labels, (
            label,
            self.labels,
        )
        dpg.set_value(self.tag, label)

    def get_next(self) -> Optional[DataSet]:
        lookup: Dict[str, DataSet] = dpg.get_item_user_data(self.tag)
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.tag)) + 1
        return lookup[labels[index % len(labels)]]

    def get_previous(self) -> Optional[DataSet]:
        lookup: Dict[str, DataSet] = dpg.get_item_user_data(self.tag)
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.tag)) - 1
        return lookup[labels[index % len(labels)]]

    def clear(self):
        dpg.configure_item(
            self.tag,
            default_value="",
        )


class ResultsCombo:
    def __init__(self, label: str, width: int):
        self.labels: List[str] = []
        dpg.add_text(label)
        self.tag: int = dpg.generate_uuid()
        dpg.add_combo(
            callback=lambda s, a, u: signals.emit(
                Signal.SELECT_TEST_RESULT,
                test=u[0].get(a),
                data=u[1],
            ),
            user_data=(
                {},
                None,
            ),
            width=width,
            tag=self.tag,
        )

    def populate(self, lookup: Dict[str, TestResult], data: Optional[DataSet]):
        self.labels.clear()
        labels: List[str] = list(lookup.keys())
        self.labels.extend(labels)
        dpg.configure_item(
            self.tag,
            items=labels,
            default_value=labels[0] if labels else "",
            user_data=(
                lookup,
                data,
            ),
        )

    def get(self) -> Optional[TestResult]:
        return dpg.get_item_user_data(self.tag)[0].get(dpg.get_value(self.tag))

    def set(self, label: str):
        assert type(label) is str, label
        assert label in self.labels, (
            label,
            self.labels,
        )
        dpg.set_value(self.tag, label)

    def clear(self):
        dpg.configure_item(
            self.tag,
            default_value="",
        )

    def get_next(self) -> Optional[TestResult]:
        lookup: Dict[str, TestResult] = dpg.get_item_user_data(self.tag)[0]
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.tag)) + 1
        return lookup[labels[index % len(labels)]]

    def get_previous(self) -> Optional[TestResult]:
        lookup: Dict[str, TestResult] = dpg.get_item_user_data(self.tag)[0]
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.tag)) - 1
        return lookup[labels[index % len(labels)]]


class KramersKronigTab:
    def __init__(self, state):
        self.state = state
        self.queued_update: Optional[Callable] = None
        self.tab: int = dpg.generate_uuid()
        label_pad: int = 24
        with dpg.tab(label="Kramers-Kronig", tag=self.tab):
            with dpg.child_window(border=False):
                with dpg.group(horizontal=True):
                    self.sidebar_window: int = dpg.generate_uuid()
                    self.sidebar_width: int = 350
                    with dpg.child_window(
                        border=False,
                        width=self.sidebar_width,
                        tag=self.sidebar_window,
                    ):
                        # TODO: Split into a separate class?
                        with dpg.child_window(width=-1, height=220):
                            self.settings_menu: SettingsMenu = SettingsMenu(
                                state.config.default_test_settings, label_pad
                            )
                            with dpg.group(horizontal=True):
                                self.visibility_item: int = dpg.generate_uuid()
                                dpg.add_text(
                                    "?".rjust(label_pad),
                                    tag=self.visibility_item,
                                )
                                attach_tooltip(tooltips.kramers_kronig.perform)
                                self.perform_test_button: int = dpg.generate_uuid()
                                dpg.add_button(
                                    label="Perform test",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.PERFORM_TEST,
                                        data=u,
                                        settings=self.get_settings(),
                                    ),
                                    user_data=None,
                                    width=-1,
                                    tag=self.perform_test_button,
                                )
                        with dpg.child_window(width=-1, height=58):
                            label_pad = 8
                            with dpg.group(horizontal=True):
                                self.data_sets_combo: DataSetsCombo = DataSetsCombo(
                                    label="Data set".rjust(label_pad),
                                    width=-60,
                                )
                            with dpg.group(horizontal=True):
                                self.results_combo: ResultsCombo = ResultsCombo(
                                    label="Result".rjust(label_pad),
                                    width=-60,
                                )
                                self.delete_button: int = dpg.generate_uuid()
                                dpg.add_button(
                                    label="Delete",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.DELETE_TEST_RESULT,
                                        **u,
                                    ),
                                    width=-1,
                                    tag=self.delete_button,
                                )
                                attach_tooltip(tooltips.kramers_kronig.delete)
                        with dpg.child_window(width=-1, height=-1):
                            with dpg.group(show=False):
                                self.validity_text: int = dpg.generate_uuid()
                                dpg.bind_item_theme(
                                    dpg.add_text(
                                        "",
                                        wrap=self.sidebar_width - 24,
                                        tag=self.validity_text,
                                    ),
                                    themes.result.invalid,
                                )
                                dpg.add_spacer(height=8)
                            self.statistics_table: StatisticsTable = StatisticsTable()
                            dpg.add_spacer(height=8)
                            self.settings_table: SettingsTable = SettingsTable()
                    self.plot_window: int = dpg.generate_uuid()
                    with dpg.child_window(border=False, tag=self.plot_window):
                        with dpg.group():
                            # Residuals
                            self.residuals_plot: Residuals = Residuals(
                                width=-1,
                                height=300,
                            )
                            self.residuals_plot.plot(
                                frequency=array([]),
                                real=array([]),
                                imaginary=array([]),
                            )
                            with dpg.group(horizontal=True):
                                self.enlarge_residuals_button: int = dpg.generate_uuid()
                                self.adjust_residuals_limits_checkbox: int = (
                                    dpg.generate_uuid()
                                )
                                dpg.add_button(
                                    label="Enlarge residuals",
                                    callback=self.show_enlarged_residuals,
                                    tag=self.enlarge_residuals_button,
                                )
                                dpg.add_checkbox(
                                    default_value=True,
                                    tag=self.adjust_residuals_limits_checkbox,
                                )
                                attach_tooltip(tooltips.general.adjust_residuals_limits)
                                dpg.add_button(
                                    label="Copy as CSV",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.COPY_PLOT_DATA,
                                        plot=self.residuals_plot,
                                        context=Context.KRAMERS_KRONIG_TAB,
                                    ),
                                )
                                attach_tooltip(tooltips.general.copy_plot_data_as_csv)
                            # Nyquist and Bode
                            self.minimum_plot_side: int = 400
                            with dpg.group(horizontal=True):
                                with dpg.group():
                                    self.nyquist_plot: Nyquist = Nyquist(
                                        width=self.minimum_plot_side,
                                        height=self.minimum_plot_side,
                                    )
                                    self.nyquist_plot.plot(
                                        real=array([]),
                                        imaginary=array([]),
                                        label="Data",
                                        line=False,
                                        theme=themes.nyquist.data,
                                    )
                                    self.nyquist_plot.plot(
                                        real=array([]),
                                        imaginary=array([]),
                                        label="Fit",
                                        line=False,
                                        fit=True,
                                        theme=themes.nyquist.simulation,
                                    )
                                    self.nyquist_plot.plot(
                                        real=array([]),
                                        imaginary=array([]),
                                        label="Fit",
                                        line=True,
                                        fit=True,
                                        theme=themes.nyquist.simulation,
                                        show_label=False,
                                    )
                                    with dpg.group(horizontal=True):
                                        self.enlarge_nyquist_button: int = (
                                            dpg.generate_uuid()
                                        )
                                        self.adjust_nyquist_limits_checkbox: int = (
                                            dpg.generate_uuid()
                                        )
                                        dpg.add_button(
                                            label="Enlarge Nyquist",
                                            callback=self.show_enlarged_nyquist,
                                            tag=self.enlarge_nyquist_button,
                                        )
                                        dpg.add_checkbox(
                                            default_value=True,
                                            tag=self.adjust_nyquist_limits_checkbox,
                                        )
                                        attach_tooltip(
                                            tooltips.general.adjust_nyquist_limits
                                        )
                                        dpg.add_button(
                                            label="Copy as CSV",
                                            callback=lambda s, a, u: signals.emit(
                                                Signal.COPY_PLOT_DATA,
                                                plot=self.nyquist_plot,
                                                context=Context.KRAMERS_KRONIG_TAB,
                                            ),
                                        )
                                        attach_tooltip(
                                            tooltips.general.copy_plot_data_as_csv
                                        )
                                self.horizontal_bode_group: int = dpg.generate_uuid()
                                with dpg.group(tag=self.horizontal_bode_group):
                                    self.bode_plot_horizontal: Bode = Bode(
                                        width=self.minimum_plot_side,
                                        height=self.minimum_plot_side,
                                    )
                                    self.bode_plot_horizontal.plot(
                                        frequency=array([]),
                                        magnitude=array([]),
                                        phase=array([]),
                                        labels=(
                                            "|Z| (d)",
                                            "phi (d)",
                                        ),
                                        line=False,
                                        themes=(
                                            themes.bode.magnitude_data,
                                            themes.bode.phase_data,
                                        ),
                                    )
                                    self.bode_plot_horizontal.plot(
                                        frequency=array([]),
                                        magnitude=array([]),
                                        phase=array([]),
                                        labels=(
                                            "|Z| (f)",
                                            "phi (f)",
                                        ),
                                        line=False,
                                        fit=True,
                                        themes=(
                                            themes.bode.magnitude_simulation,
                                            themes.bode.phase_simulation,
                                        ),
                                    )
                                    self.bode_plot_horizontal.plot(
                                        frequency=array([]),
                                        magnitude=array([]),
                                        phase=array([]),
                                        labels=(
                                            "|Z| (f)",
                                            "phi (f)",
                                        ),
                                        line=True,
                                        fit=True,
                                        themes=(
                                            themes.bode.magnitude_simulation,
                                            themes.bode.phase_simulation,
                                        ),
                                        show_labels=False,
                                    )
                                    with dpg.group(horizontal=True):
                                        self.enlarge_bode_horizontal_button: int = (
                                            dpg.generate_uuid()
                                        )
                                        self.adjust_bode_limits_horizontal_checkbox: int = (
                                            dpg.generate_uuid()
                                        )
                                        dpg.add_button(
                                            label="Enlarge Bode",
                                            callback=self.show_enlarged_bode,
                                            tag=self.enlarge_bode_horizontal_button,
                                        )
                                        dpg.add_checkbox(
                                            default_value=True,
                                            tag=self.adjust_bode_limits_horizontal_checkbox,
                                        )
                                        attach_tooltip(
                                            tooltips.general.adjust_bode_limits
                                        )
                                        dpg.add_button(
                                            label="Copy as CSV",
                                            callback=lambda s, a, u: signals.emit(
                                                Signal.COPY_PLOT_DATA,
                                                plot=self.bode_plot_horizontal,
                                                context=Context.KRAMERS_KRONIG_TAB,
                                            ),
                                        )
                                        attach_tooltip(
                                            tooltips.general.copy_plot_data_as_csv
                                        )
                            self.vertical_bode_group: int = dpg.generate_uuid()
                            with dpg.group(tag=self.vertical_bode_group):
                                self.bode_plot_vertical: Bode = Bode(
                                    width=-1,
                                    height=self.minimum_plot_side,
                                )
                                self.bode_plot_vertical.plot(
                                    frequency=array([]),
                                    magnitude=array([]),
                                    phase=array([]),
                                    labels=(
                                        "|Z| (d)",
                                        "phi (d)",
                                    ),
                                    line=False,
                                    themes=(
                                        themes.bode.magnitude_data,
                                        themes.bode.phase_data,
                                    ),
                                )
                                self.bode_plot_vertical.plot(
                                    frequency=array([]),
                                    magnitude=array([]),
                                    phase=array([]),
                                    labels=(
                                        "|Z| (f)",
                                        "phi (f)",
                                    ),
                                    line=False,
                                    fit=True,
                                    themes=(
                                        themes.bode.magnitude_simulation,
                                        themes.bode.phase_simulation,
                                    ),
                                )
                                self.bode_plot_vertical.plot(
                                    frequency=array([]),
                                    magnitude=array([]),
                                    phase=array([]),
                                    labels=(
                                        "|Z| (f)",
                                        "phi (f)",
                                    ),
                                    line=True,
                                    fit=True,
                                    themes=(
                                        themes.bode.magnitude_simulation,
                                        themes.bode.phase_simulation,
                                    ),
                                    show_labels=False,
                                )
                                with dpg.group(horizontal=True):
                                    self.enlarge_bode_vertical_button: int = (
                                        dpg.generate_uuid()
                                    )
                                    self.adjust_bode_limits_vertical_checkbox: int = (
                                        dpg.generate_uuid()
                                    )
                                    dpg.add_button(
                                        label="Enlarge Bode",
                                        callback=lambda s, a, u: signals.emit(
                                            Signal.SHOW_ENLARGED_PLOT,
                                            plot=self.bode_plot_vertical,
                                            adjust_limits=dpg.get_value(
                                                self.adjust_bode_limits_horizontal_checkbox
                                            ),
                                        ),
                                        tag=self.enlarge_bode_vertical_button,
                                    )
                                    dpg.add_checkbox(
                                        default_value=True,
                                        source=self.adjust_bode_limits_horizontal_checkbox,
                                        tag=self.adjust_bode_limits_vertical_checkbox,
                                    )
                                    attach_tooltip(tooltips.general.adjust_bode_limits)
                                    dpg.add_button(
                                        label="Copy as CSV",
                                        callback=lambda s, a, u: signals.emit(
                                            Signal.COPY_PLOT_DATA,
                                            plot=self.bode_plot_vertical,
                                            context=Context.KRAMERS_KRONIG_TAB,
                                        ),
                                    )
                                    attach_tooltip(
                                        tooltips.general.copy_plot_data_as_csv
                                    )
        self.set_settings(self.state.config.default_test_settings)

    def is_visible(self) -> bool:
        return dpg.is_item_visible(self.visibility_item)

    def resize(self, width: int, height: int):
        assert type(width) is int and width > 0
        assert type(height) is int and height > 0
        if not self.is_visible():
            return
        if width < (self.sidebar_width + self.minimum_plot_side * 2):
            if dpg.is_item_shown(self.horizontal_bode_group):
                dpg.hide_item(self.horizontal_bode_group)
                dpg.show_item(self.vertical_bode_group)
                self.nyquist_plot.resize(-1, self.minimum_plot_side)
        else:
            if dpg.is_item_shown(self.vertical_bode_group):
                dpg.show_item(self.horizontal_bode_group)
                dpg.hide_item(self.vertical_bode_group)
            dpg.split_frame()
            width, height = dpg.get_item_rect_size(self.plot_window)
            width = round((width - 24) / 2) + 7
            height = height - 300 - 24 * 2 - 2
            self.nyquist_plot.resize(width, height)
            self.bode_plot_horizontal.resize(width, height)

    def clear(self, hide: bool = True):
        self.data_sets_combo.clear()
        self.results_combo.clear()
        self.statistics_table.clear(hide=hide)
        self.settings_table.clear(hide=hide)
        dpg.set_item_user_data(self.perform_test_button, None)
        self.residuals_plot.clear(delete=False)
        self.nyquist_plot.clear(delete=False)
        self.bode_plot_horizontal.clear(delete=False)
        self.bode_plot_vertical.clear(delete=False)

    def get_settings(self) -> TestSettings:
        return self.settings_menu.get_settings()

    def set_settings(self, settings: TestSettings):
        self.settings_menu.set_settings(settings)

    def populate_data_sets(self, labels: List[str], lookup: Dict[str, DataSet]):
        assert type(labels) is list, labels
        assert type(lookup) is dict, lookup
        self.data_sets_combo.populate(labels, lookup)

    def populate_tests(self, lookup: Dict[str, TestResult], data: Optional[DataSet]):
        assert type(lookup) is dict, lookup
        assert type(data) is DataSet or data is None, data
        self.results_combo.populate(lookup, data)
        dpg.hide_item(dpg.get_item_parent(self.validity_text))
        if data is not None and self.results_combo.labels:
            signals.emit(
                Signal.SELECT_TEST_RESULT,
                test=self.results_combo.get(),
                data=data,
            )
        else:
            self.statistics_table.clear(hide=True)
            self.settings_table.clear(hide=True)
            self.select_data_set(data)

    def get_next_data_set(self) -> Optional[DataSet]:
        return self.data_sets_combo.get_next()

    def get_previous_data_set(self) -> Optional[DataSet]:
        return self.data_sets_combo.get_previous()

    def get_next_result(self) -> Optional[TestResult]:
        return self.results_combo.get_next()

    def get_previous_result(self) -> Optional[TestResult]:
        return self.results_combo.get_previous()

    def select_data_set(self, data: Optional[DataSet]):
        assert type(data) is DataSet or data is None, data
        self.clear(hide=data is None)
        dpg.set_item_user_data(self.perform_test_button, data)
        self.settings_menu.update_num_RC_slider(data)
        if data is None:
            return
        self.data_sets_combo.set(data.get_label())
        real: ndarray
        imag: ndarray
        real, imag = data.get_nyquist_data()
        self.nyquist_plot.update(
            index=0,
            real=real,
            imaginary=imag,
        )
        freq: ndarray
        mag: ndarray
        phase: ndarray
        freq, mag, phase = data.get_bode_data()
        self.bode_plot_horizontal.update(
            index=0,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        self.bode_plot_vertical.update(
            index=0,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )

    def assert_test_up_to_date(self, test: TestResult, data: DataSet):
        # Check if the number of unmasked points is the same
        Z_exp: ndarray = data.get_impedance()
        Z_test: ndarray = test.get_impedance()
        assert Z_exp.shape == Z_test.shape, "The number of data points differ!"
        # Check if the masks are the same
        mask_exp: Dict[int, bool] = data.get_mask()
        mask_test: Dict[int, bool] = {
            k: test.mask.get(k, mask_exp.get(k, False)) for k in test.mask
        }
        num_masked_exp: int = list(data.get_mask().values()).count(True)
        num_masked_test: int = list(test.mask.values()).count(True)
        assert num_masked_exp == num_masked_test, "The masks are different sizes!"
        i: int
        for i in mask_test.keys():
            assert (
                i in mask_exp
            ), f"The data set does not have a point at index {i + 1}!"
            assert (
                mask_exp[i] == mask_test[i]
            ), f"The data set's mask differs at index {i + 1}!"
        # Check if the frequencies and impedances are the same
        assert allclose(
            test.get_frequency(), data.get_frequency()
        ), "The frequencies differ!"
        real_residual: ndarray
        imaginary_residual: ndarray
        real_residual, imaginary_residual = _calculate_residuals(Z_exp, Z_test)
        assert allclose(test.real_residual, real_residual) and allclose(
            test.imaginary_residual, imaginary_residual
        ), "The data set's impedances differ from what they were when the test was performed!"

    def select_test_result(self, test: Optional[TestResult], data: Optional[DataSet]):
        assert type(test) is TestResult or test is None, test
        assert type(data) is DataSet or data is None, data
        dpg.set_item_user_data(
            self.delete_button,
            {
                "test": test,
                "data": data,
            },
        )
        if not self.is_visible():
            self.queued_update = lambda: self.select_test_result(test, data)
            return
        self.queued_update = None
        self.select_data_set(data)
        if test is None or data is None:
            if dpg.get_value(self.adjust_residuals_limits_checkbox):
                self.residuals_plot.queue_limits_adjustment()
            if dpg.get_value(self.adjust_nyquist_limits_checkbox):
                self.nyquist_plot.queue_limits_adjustment()
            if dpg.get_value(self.adjust_bode_limits_horizontal_checkbox):
                self.bode_plot_horizontal.queue_limits_adjustment()
                self.bode_plot_vertical.queue_limits_adjustment()
            return
        self.results_combo.set(test.get_label())
        message: str
        try:
            self.assert_test_up_to_date(test, data)
            dpg.hide_item(dpg.get_item_parent(self.validity_text))
        except AssertionError as message:
            dpg.set_value(
                self.validity_text,
                f"Test result is not valid for the current state of the data set!\n\n{message}",
            )
            dpg.show_item(dpg.get_item_parent(self.validity_text))
        self.statistics_table.populate(test)
        self.settings_table.populate(
            test,
            data,
            self.settings_menu.get_num_RC_labels(),
        )
        freq: ndarray
        real: ndarray
        imag: ndarray
        freq, real, imag = test.get_residual_data()
        self.residuals_plot.update(
            index=0,
            frequency=freq,
            real=real,
            imaginary=imag,
        )
        real, imag = test.get_nyquist_data()
        self.nyquist_plot.update(
            index=1,
            real=real,
            imaginary=imag,
        )
        real, imag = test.get_nyquist_data(
            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
        )
        self.nyquist_plot.update(
            index=2,
            real=real,
            imaginary=imag,
        )
        mag: ndarray
        phase: ndarray
        freq, mag, phase = test.get_bode_data()
        self.bode_plot_horizontal.update(
            index=1,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        freq, mag, phase = test.get_bode_data(
            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
        )
        self.bode_plot_horizontal.update(
            index=2,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        freq, mag, phase = test.get_bode_data()
        self.bode_plot_vertical.update(
            index=1,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        freq, mag, phase = test.get_bode_data(
            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
        )
        self.bode_plot_vertical.update(
            index=2,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        if dpg.get_value(self.adjust_residuals_limits_checkbox):
            self.residuals_plot.queue_limits_adjustment()
        if dpg.get_value(self.adjust_nyquist_limits_checkbox):
            self.nyquist_plot.queue_limits_adjustment()
        if dpg.get_value(self.adjust_bode_limits_horizontal_checkbox):
            self.bode_plot_horizontal.queue_limits_adjustment()
            self.bode_plot_vertical.queue_limits_adjustment()

    def show_enlarged_nyquist(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.nyquist_plot,
            adjust_limits=dpg.get_value(self.adjust_nyquist_limits_checkbox),
        )

    def show_enlarged_bode(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.bode_plot_horizontal,
            adjust_limits=dpg.get_value(self.adjust_bode_limits_horizontal_checkbox),
        )

    def show_enlarged_residuals(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.residuals_plot,
            adjust_limits=dpg.get_value(self.adjust_residuals_limits_checkbox),
        )

    def has_active_input(self) -> bool:
        return self.settings_menu.has_active_input()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import List, Optional
import dearpygui.dearpygui as dpg


class Plot:
    def __init__(self):
        self._plot: int = dpg.generate_uuid()
        self._series: List[dict] = []
        self._item_handler: int = dpg.generate_uuid()
        with dpg.item_handler_registry(tag=self._item_handler):
            dpg.add_item_visible_handler(callback=self._visibility_handler)
        self._is_limits_adjustment_queued: bool = False

    def _visibility_handler(self, sender, app):
        if not self._is_limits_adjustment_queued:
            return
        dpg.split_frame()
        self.adjust_limits()
        self._is_limits_adjustment_queued = False

    @classmethod
    def duplicate(Class, original: "Plot", *args, **kwargs) -> "Plot":
        raise Exception("Method has not been implemented!")

    def copy_limits(self, other: "Plot"):
        raise Exception("Method has not been implemented!")

    def show(self):
        dpg.show_item(self._plot)

    def hide(self):
        dpg.hide_item(self._plot)

    def delete(self):
        dpg.delete_item(self._item_handler)
        dpg.delete_item(self._plot)

    def set_title(self, title: Optional[str]):
        dpg.set_item_label(self._plot, title)

    def are_limits_adjusted(self) -> bool:
        return not self._is_limits_adjustment_queued

    def queue_limits_adjustment(self):
        self._is_limits_adjustment_queued = True

    def limits_adjusted(self):
        self._is_limits_adjustment_queued = False

    def resize(self, width: int = -1, height: int = -1):
        assert type(width) is int, width
        assert type(height) is int, height
        dpg.configure_item(self._plot, width=width, height=height)

    def get_series(self) -> List[dict]:
        return self._series.copy()

    def is_visible(self) -> bool:
        return dpg.is_item_visible(self._plot)

    def clear(self, *args, **kwargs):
        raise Exception("'clear' method has not been implemented!")

    def plot(self, *args, **kwargs):
        raise Exception("'plot' method has not been implemented!")



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from types import SimpleNamespace

kulikovsky2020: str = "DOI:10.1039/D0CP02094J"
wan2015: str = "DOI:10.1016/j.electacta.2015.09.097"
ciucci2015: str = "DOI:10.1016/j.electacta.2015.03.123"
effat2017: str = "DOI:10.1016/j.electacta.2017.07.050"
liu2020: str = "DOI:10.1016/j.electacta.2020.136864"
boukamp2015: str = "DOI:10.1016/j.electacta.2014.12.059"
boukamp2017: str = "DOI:10.1016/j.ssi.2016.10.009"

drt = SimpleNamespace(
    **{
        "method": f"""
The method to use when calculating the distribution of relaxation times.

BHT: Bayesian Hilbert transform method.

TR-NNLS: Tikhonov regularization combined with non-negative least squares fitting.

TR-RBF: Tikhonov regularization combined with radial basis function (or piecewise linear) discretization and convex optimization.

m(RQ)fit: multi-(RQ) complex non-linear least squares fit.

The TR-NNLS method is very fast while the slower TR-RBF method provides the ability to optionally also calculate Bayesian credible intervals. The BHT method, while temperamental due to randomization of initial values, is able to provide numerical scores that can be used to assess the quality of an impedance spectrum. The m(RQ)fit method requires a suitable fitted circuit but may be able to distinguish peaks that are in close proximity and would only show up as broad peaks in some other methods.

References:
- {kulikovsky2020}
- {wan2015}
- {ciucci2015}
- {effat2017}
- {liu2020}
- {boukamp2015}
- {boukamp2017}
    """.strip(),
        "mode": """
The data to use when performing the calculations:
- Complex
- Real
- Imaginary

This is only used when the method setting is set to TR-NNLS or TR-RBF.
    """.strip(),
        "lambda_value": """
The regularization parameter to use as part of the Tikhonov regularization. If the checkbox next to the input field is ticked, then an attempt will be made to automatically find a suitable value. However, further tweaking of the value manually is recommended.

This is only used when the method setting is set to TR-NNLS or TR-RBF.
    """.strip(),
        "shape_coeff": """
The input value, k, for the shape coefficient used in the shape type (see the setting above).

This is only used when the method setting is set to BHT or TR-RBF.
    """.strip(),
        "credible_intervals": """
Whether or not the Bayesican credible intervals are calculated when using the TR-RBF method. The number of samples to use can also be defined. More accurate results can be obtained with a greater number of samples albeit at the expense of requiring more time.
    """.strip(),
        "num_samples": """
The number of samples to use when:
- calculating Bayesian credible intervals (TR-RBF method)
- calculating Jensen-Shannon distances (BHT method)

A greater number of samples provide greater accuracy at the expense of time.
    """.strip(),
        "num_attempts": """
The number of attempts to make to find a solution using the BHT method. The initial values are randomized so a greater number of attempts is more likely to yield reasonable results at the expense of requiring more time. If poor results (e.g., significant oscillation in the DRT plot) are obtained, then try repeating the calculations and possibly also tweak the settings.
    """.strip(),
        "maximum_symmetry": """
Poor results with significant oscillation can be discarded automatically by defining a limit for the ratio of the vertical peak-to-peak symmetry allowed in the DRT plot. The limit is defined as a ratio from 0.0 to 1.0. A smaller value provides a stricter condition as the absolute value of the most positive peak must be greater than the absolute value of the most negative peak.

This is only used when the method setting is set to BHT or TR-RBF.
    """.strip(),
        "perform": f"""
DRT analyses work best with data sets where the imaginary part of the impedance approaches zero at both the low- and high-frequency ends of the recorded frequency range. Thus, some data sets may require some processing if they exhibit, e.g., inductive behavior at high frequencies and/or increasing impedance at low frequencies due to diffusion.

If the BHT or TR-RBF methods are used for the first time, then keeping the number of samples and attempts modest at first is recommended. It may take quite a long time depending on the computer's performance.

If the BHT method is used, then scores for the consistency of the data will be included. The scores range from 0 to 100 %.

Reference: {liu2020}
    """.strip(),
        "delete": """
Delete the current analysis result.
    """.strip(),
        "score_mean": """
The similarity of the means of the Hilbert transformed impedances and the DRT impedances (i.e., without the high-frequency resistance and inductance).
    """.strip(),
        "score_residuals": """
The amount of experimental points that fall within one, two, or three standard deviations of the prediction of the Hilbert transform.
    """.strip(),
        "score_hd": """
The Hellinger distance between the Hilbert transformed impedances and the DRT impedances (i.e., without the high-frequency resistance and inductance).
    """.strip(),
        "score_jsd": """
The Jensen-Shannon distance between the Hilbert transformed impedances and the DRT impedances (i.e., without the high-frequency resistance and inductance).
    """.strip(),
        "circuit": """
The fitted circuit to use to calculate the distribution of relaxation times.

Valid circuits consist of one or more parallel RQ (or RC) circuits in series. An optional series resistance may also be included. For example, the following circuits would be valid:
- (RC)
- (RQ)
- (RQ)(RQ)
- R(RQ)
- R(RQ)(RQ)(RC)

This is only used when the method is set to m(RQ)fit.
    """.strip(),
        "W": """
The width of the Gauss function used to calculate the distribution of relaxation times for (RC) circuits (or (RQ) circuits where n ~ 1).

This is only used when the method is set to m(RQ)fit.
    """.strip(),
        "num_per_decade": """
The number of points per decade to use when calculating the distribution of relaxation times. This only affects the final step where the distribution of relaxation times are calculated using the analytical solutions (or approximations in the case of (RC) circuits).

This is only used when the method is set to m(RQ)fit.
    """.strip(),
    }
)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import Callable, List, Optional
import dearpygui.dearpygui as dpg
from numpy import array, ndarray
import deareis.themes as themes
from deareis.gui.plots.base import Plot


class Nyquist(Plot):
    def __init__(self, width: int = -1, height: int = -1, *args, **kwargs):
        assert type(width) is int, width
        assert type(height) is int, height
        super().__init__()
        with dpg.plot(
            anti_aliased=True,
            crosshairs=True,
            equal_aspects=True,
            width=width,
            height=height,
            tag=self._plot,
        ):
            dpg.add_plot_legend(
                horizontal=kwargs.get("legend_horizontal", True),
                location=kwargs.get("legend_location", dpg.mvPlot_Location_North),
                outside=kwargs.get("legend_outside", True),
            )
            self._x_axis: int = dpg.add_plot_axis(
                dpg.mvXAxis,
                label="Z' (ohm)",
                no_gridlines=True,
            )
            self._y_axis: int = dpg.add_plot_axis(
                dpg.mvYAxis,
                label='-Z" (ohm)',
                no_gridlines=True,
            )
        dpg.bind_item_theme(self._plot, themes.plot)
        dpg.bind_item_handler_registry(self._plot, self._item_handler)

    @classmethod
    def duplicate(Class, original: Plot, *args, **kwargs) -> Plot:
        copy: Plot = Class(*args, **kwargs)
        for kwargs in original.get_series():
            copy.plot(**kwargs)
        return copy

    def is_blank(self) -> bool:
        return len(dpg.get_item_children(self._y_axis, slot=1)) == 0

    def clear(self, *args, **kwargs):
        delete: bool = kwargs.get("delete", True)
        if delete:
            dpg.delete_item(self._y_axis, children_only=True)
            self._series.clear()
        else:
            i: int
            series: int
            for i, series in enumerate(dpg.get_item_children(self._y_axis, slot=1)):
                self._series[i]["real"] = array([])
                self._series[i]["imaginary"] = array([])
                dpg.set_value(series, [[], []])

    def update(self, index: int, *args, **kwargs):
        assert type(index) is int and index >= 0, index
        assert len(self._series) > index, (
            index,
            len(self._series),
        )
        assert len(args) == 0, args
        real: ndarray = kwargs["real"]
        imag: ndarray = kwargs["imaginary"]
        assert type(real) is ndarray, real
        assert type(imag) is ndarray, imag
        i: int
        series: int
        for i, series in enumerate(dpg.get_item_children(self._y_axis, slot=1)):
            if i != index:
                continue
            self._series[index].update(kwargs)
            dpg.set_value(series, [list(real), list(imag)])
            break

    def plot(self, *args, **kwargs) -> int:
        assert len(args) == 0, args
        real: ndarray = kwargs["real"]
        imag: ndarray = kwargs["imaginary"]
        label: str = kwargs["label"]
        simulation: bool = kwargs.get("simulation", False)
        fit: bool = kwargs.get("fit", False)
        line: bool = kwargs.get("line", False)
        theme: Optional[int] = kwargs.get("theme")
        show_label: bool = kwargs.get("show_label", True)
        assert type(real) is ndarray, real
        assert type(imag) is ndarray, imag
        assert type(label) is str, label
        assert type(simulation) is bool, simulation
        assert type(fit) is bool, fit
        assert type(line) is bool, line
        assert type(theme) is int or theme is None, theme
        assert type(show_label) is bool, show_label
        self._series.append(kwargs)
        func: Callable = dpg.add_scatter_series if not line else dpg.add_line_series
        tag: int = func(
            x=list(real),
            y=list(imag),
            label=label if show_label else None,
            parent=self._y_axis,
        )
        if theme is not None:
            dpg.bind_item_theme(tag, theme)
        return tag

    def adjust_limits(self):
        if not self.is_visible():
            self.queue_limits_adjustment()
            return
        elif self.are_limits_adjusted():
            return
        else:
            self.limits_adjusted()
        dpg.split_frame()
        dpg.fit_axis_data(self._x_axis)
        dpg.fit_axis_data(self._y_axis)
        dpg.split_frame()
        x_min: float
        x_max: float
        x_min, x_max = dpg.get_axis_limits(self._x_axis)
        dx: float = (x_max - x_min) * 0.05
        dpg.set_axis_limits(self._x_axis, x_min - dx, x_max + dx)
        y_min: float
        y_max: float
        y_min, y_max = dpg.get_axis_limits(self._y_axis)
        dy: float = (y_max - y_min) * 0.05
        dpg.set_axis_limits(self._y_axis, y_min - dy, y_max + dy)
        dpg.split_frame()
        dpg.set_axis_limits_auto(self._x_axis)
        dpg.set_axis_limits_auto(self._y_axis)

    def copy_limits(self, other: Plot):
        src: int
        dst: int
        for src, dst in zip(
            [
                other._x_axis,
                other._y_axis,
            ],
            [
                self._x_axis,
                self._y_axis,
            ],
        ):
            limits: List[float] = dpg.get_axis_limits(src)
            dpg.set_axis_limits(dst, *limits)
        dpg.split_frame()
        dpg.set_axis_limits_auto(self._x_axis)
        dpg.set_axis_limits_auto(self._y_axis)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

import dearpygui.dearpygui as dpg
from .general import general
from .data_sets import data_sets
from .kramers_kronig import kramers_kronig
from .fitting import fitting
from .simulation import simulation
from .plotting import plotting
from .circuit_editor import circuit_editor
from .recent_projects import recent_projects
from .home import home
from .drt import drt


def update_tooltip(tag: int, msg: str, wrap: bool = True):
    assert type(tag) is int, tag
    assert type(msg) is str, msg
    assert type(wrap) is bool, wrap
    wrap_limit: int = -1
    if wrap:
        max_line_length: int
        if "\n" in msg:
            max_line_length = max(map(len, msg.split("\n")))
        else:
            max_line_length = len(msg)
        wrap_limit = min([8 * max_line_length, 500])
    dpg.configure_item(tag, default_value=msg, wrap=wrap_limit)


def attach_tooltip(msg: str, parent: int = -1, tag: int = -1, wrap: bool = True) -> int:
    assert type(msg) is str, msg
    assert type(parent) is int, parent
    assert type(tag) is int, tag
    assert type(wrap) is bool, wrap
    if parent < 0:
        parent = dpg.last_item()
    if tag < 0:
        tag = dpg.generate_uuid()
    with dpg.tooltip(parent, user_data=tag):
        dpg.add_text("", tag=tag)
    update_tooltip(tag, msg, wrap)
    return tag



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import Dict
from dataclasses import dataclass
import dearpygui.dearpygui as dpg
from deareis.enums import Action, action_to_string, string_to_action


dpg_to_string: Dict[int, str] = {
    dpg.mvKey_A: "A",
    dpg.mvKey_B: "B",
    dpg.mvKey_C: "C",
    dpg.mvKey_D: "D",
    dpg.mvKey_E: "E",
    dpg.mvKey_F: "F",
    dpg.mvKey_G: "G",
    dpg.mvKey_H: "H",
    dpg.mvKey_I: "I",
    dpg.mvKey_J: "J",
    dpg.mvKey_K: "K",
    dpg.mvKey_L: "L",
    dpg.mvKey_M: "M",
    dpg.mvKey_N: "N",
    dpg.mvKey_O: "O",
    dpg.mvKey_P: "P",
    dpg.mvKey_Q: "Q",
    dpg.mvKey_R: "R",
    dpg.mvKey_S: "S",
    dpg.mvKey_T: "T",
    dpg.mvKey_U: "U",
    dpg.mvKey_V: "V",
    dpg.mvKey_W: "W",
    dpg.mvKey_X: "X",
    dpg.mvKey_Y: "Y",
    dpg.mvKey_Z: "Z",
    dpg.mvKey_0: "0",
    dpg.mvKey_1: "1",
    dpg.mvKey_2: "2",
    dpg.mvKey_3: "3",
    dpg.mvKey_4: "4",
    dpg.mvKey_5: "5",
    dpg.mvKey_6: "6",
    dpg.mvKey_7: "7",
    dpg.mvKey_8: "8",
    dpg.mvKey_9: "9",
    # dpg.mvKey_Back: "",Back
    dpg.mvKey_Tab: "Tab",
    # dpg.mvKey_Clear: "",Clear
    dpg.mvKey_Return: "Enter",
    # dpg.mvKey_Shift: "Shift",
    # dpg.mvKey_Control: "Control",
    # dpg.mvKey_Alt: "",Alt
    # dpg.mvKey_Pause: "",Pause
    # dpg.mvKey_Capital: "",Capital
    dpg.mvKey_Escape: "Esc",
    dpg.mvKey_Spacebar: "Spacebar",
    dpg.mvKey_Prior: "Page up",
    dpg.mvKey_Next: "Page down",
    dpg.mvKey_End: "End",
    dpg.mvKey_Home: "Home",
    dpg.mvKey_Left: "Left",
    dpg.mvKey_Up: "Up",
    dpg.mvKey_Right: "Right",
    dpg.mvKey_Down: "Down",
    # dpg.mvKey_Select: "",Select
    # dpg.mvKey_Print: "",Print
    # dpg.mvKey_Execute: "",Execute
    # dpg.mvKey_PrintScreen: "Print screen",
    dpg.mvKey_Insert: "Insert",
    dpg.mvKey_Delete: "Delete",
    # dpg.mvKey_Help: "",Help
    # dpg.mvKey_LWin: "",LWin
    # dpg.mvKey_RWin: "",RWin
    # dpg.mvKey_Apps: "",Apps
    # dpg.mvKey_Sleep: "",Sleep
    dpg.mvKey_NumPad0: "Numpad 0",
    dpg.mvKey_NumPad1: "Numpad 1",
    dpg.mvKey_NumPad2: "Numpad 2",
    dpg.mvKey_NumPad3: "Numpad 3",
    dpg.mvKey_NumPad4: "Numpad 4",
    dpg.mvKey_NumPad5: "Numpad 5",
    dpg.mvKey_NumPad6: "Numpad 6",
    dpg.mvKey_NumPad7: "Numpad 7",
    dpg.mvKey_NumPad8: "Numpad 8",
    dpg.mvKey_NumPad9: "Numpad 9",
    # dpg.mvKey_Multiply: "",Multiply
    # dpg.mvKey_Add: "",Add
    # dpg.mvKey_Separator: "",Separator
    # dpg.mvKey_Subtract: "",Subtract
    # dpg.mvKey_Decimal: "",Decimal
    # dpg.mvKey_Divide: "",Divide
    dpg.mvKey_F1: "F1",
    dpg.mvKey_F2: "F2",
    dpg.mvKey_F3: "F3",
    dpg.mvKey_F4: "F4",
    dpg.mvKey_F5: "F5",
    dpg.mvKey_F6: "F6",
    dpg.mvKey_F7: "F7",
    dpg.mvKey_F8: "F8",
    dpg.mvKey_F9: "F9",
    dpg.mvKey_F10: "F10",
    dpg.mvKey_F11: "F11",
    dpg.mvKey_F12: "F12",
    dpg.mvKey_F13: "F13",
    dpg.mvKey_F14: "F14",
    dpg.mvKey_F15: "F15",
    dpg.mvKey_F16: "F16",
    dpg.mvKey_F17: "F17",
    dpg.mvKey_F18: "F18",
    dpg.mvKey_F19: "F19",
    dpg.mvKey_F20: "F20",
    dpg.mvKey_F21: "F21",
    dpg.mvKey_F22: "F22",
    dpg.mvKey_F23: "F23",
    dpg.mvKey_F24: "F24",
    dpg.mvKey_F25: "F25",
    dpg.mvKey_NumLock: "Num lock",
    dpg.mvKey_ScrollLock: "Scroll lock",
    dpg.mvKey_LShift: "Shift (left)",
    dpg.mvKey_RShift: "Shift (right)",
    dpg.mvKey_LControl: "Control (left)",
    dpg.mvKey_RControl: "Control (right)",
    # dpg.mvKey_LMenu: "",LMenu
    # dpg.mvKey_RMenu: "",RMenu
    # dpg.mvKey_Browser_Back: "",Browser_Back
    # dpg.mvKey_Browser_Forward: "",Browser_Forward
    # dpg.mvKey_Browser_Refresh: "",Browser_Refresh
    # dpg.mvKey_Browser_Stop: "",Browser_Stop
    # dpg.mvKey_Browser_Search: "",Browser_Search
    # dpg.mvKey_Browser_Favorites: "",Browser_Favorites
    # dpg.mvKey_Browser_Home: "",Browser_Home
    # dpg.mvKey_Volume_Mute: "",Volume_Mute
    # dpg.mvKey_Volume_Down: "",Volume_Down
    # dpg.mvKey_Volume_Up: "",Volume_Up
    # dpg.mvKey_Media_Next_Track: "",Media_Next_Track
    # dpg.mvKey_Media_Prev_Track: "",Media_Prev_Track
    # dpg.mvKey_Media_Stop: "",Media_Stop
    # dpg.mvKey_Media_Play_Pause: "",Media_Play_Pause
    # dpg.mvKey_Launch_Mail: "",Launch_Mail
    # dpg.mvKey_Launch_Media_Select: "",Launch_Media_Select
    # dpg.mvKey_Launch_App1: "",Launch_App1
    # dpg.mvKey_Launch_App2: "",Launch_App2
    # dpg.mvKey_Colon: ":",
    # dpg.mvKey_Plus: "+",
    # dpg.mvKey_Comma: ",",
    # dpg.mvKey_Minus: "-",
    # dpg.mvKey_Period: ".",
    # dpg.mvKey_Slash: "/",
    # dpg.mvKey_Tilde: "~",
    # dpg.mvKey_Open_Brace: "",Open_Brace
    # dpg.mvKey_Backslash: "\\",
    # dpg.mvKey_Close_Brace: "",Close_Brace
    # dpg.mvKey_Quote: "",Quote
    # dpg.mvKey_: "",
}
string_to_dpg: Dict[str, int] = {v: k for k, v in dpg_to_string.items()}


@dataclass
class Keybinding:
    key: int
    mod_alt: bool
    mod_ctrl: bool
    mod_shift: bool
    action: Action

    def __contains__(self, key: int) -> bool:
        assert key in dpg_to_string, key
        if self.key == key:
            return True
        elif self.mod_alt and key == dpg.mvKey_Alt:
            return True
        elif self.mod_ctrl and (
            key == dpg.mvKey_Control
            or key == dpg.mvKey_LControl
            or key == dpg.mvKey_RControl
        ):
            return True
        elif self.mod_shift and (key == dpg.mvKey_Shift or key == dpg.mvKey_RShift):
            return True
        return False

    def __str__(self) -> str:
        string: str = self.__repr__()
        return string[: string.find(":")]

    def __repr__(self) -> str:
        string: str = dpg_to_string[self.key]
        if self.mod_shift:
            string = f"Shift+{string}"
        if self.mod_ctrl:
            string = f"Ctrl+{string}"
        if self.mod_alt:
            string = f"Alt+{string}"
        string = f"{string}: {action_to_string[self.action]}"
        return string

    def to_dict(self) -> dict:
        return {
            "key": dpg_to_string[self.key],
            "mod_alt": self.mod_alt,
            "mod_ctrl": self.mod_ctrl,
            "mod_shift": self.mod_shift,
            "action": action_to_string[self.action],
        }

    @classmethod
    def from_dict(Class, dictionary) -> "Keybinding":
        dictionary["key"] = string_to_dpg[dictionary["key"]]
        dictionary["action"] = string_to_action[dictionary["action"]]
        return Class(**dictionary)
